import json
from pathlib import Path
from typing import Any, Callable, Optional

from fused.models._codegen import MetaJson
from fused.models.api import UdfJobStepConfig
from fused.models.udf._eval_result import UdfEvaluationResult
from fused.models.udf.output import Output
from fused.models.udf.udf import GeoPandasUdfV2, load_udf_from_response_data


def get_step_config_from_server(
    email_or_handle: Optional[str],
    slug: str,
    cache_key: Any,
    _is_public: bool = False,
) -> UdfJobStepConfig:
    from fused.api.api import FusedAPI

    # cache_key is unused
    api = FusedAPI()
    if _is_public:
        obj = api._get_public_udf(slug)
    else:
        obj = api._get_udf(email_or_handle, slug)
    udf = load_udf_from_response_data(obj)

    step_config = UdfJobStepConfig(udf=udf)
    return step_config


def get_github_udf_from_server(
    url: str,
    cache_key: Any,
):
    from fused.api.api import FusedAPI

    # cache_key is unused
    # TODO: Do this locally in fused-py
    api = FusedAPI(credentials_needed=False)
    obj = api._get_code_by_url(url)
    udf = load_udf_from_response_data(obj)

    step_config = UdfJobStepConfig(udf=udf)
    return step_config


def run_and_get_data(udf, *args, **kwargs):
    # TODO: This is a silly way to do this, because we have to pass parameters in such an odd way
    job = udf(*args, **kwargs)
    result = job.run_local()
    if isinstance(result, (Output, UdfEvaluationResult)):
        return result.data
    else:
        return result


def get_step_config_from_shared_token(token: str) -> UdfJobStepConfig:
    from fused.api.api import FusedAPI

    api = FusedAPI()
    obj = api._get_udf_by_token(token)
    udf = load_udf_from_response_data(obj)
    return UdfJobStepConfig(udf=udf)


def get_udf_from_file(path: Path) -> GeoPandasUdfV2:
    data = {
        "name": path.stem,
        "entrypoint": "udf",
        "type": "geopandas_v2",
        "code": path.read_bytes().decode("utf8"),
    }
    return GeoPandasUdfV2.model_validate(data)


def get_udf_from_code(code: str, name: Optional[str] = None) -> GeoPandasUdfV2:
    data = {
        "name": name or "udf",
        "entrypoint": "udf",
        "type": "geopandas_v2",
        "code": code,
    }
    return GeoPandasUdfV2.model_validate(data)


def get_udf_from_directory(load_callback: Callable[[str], bytes]) -> GeoPandasUdfV2:
    meta_contents = json.loads(load_callback("meta.json"))
    meta = MetaJson.model_validate(meta_contents)

    if len(meta.job_config.steps) != 1:
        raise ValueError(
            f"meta.json is not in expected format: {len(meta.job_config.steps)=}"
        )

    if meta.job_config.steps[0]["type"] != "udf":
        raise ValueError(
            f'meta.json is not in expected format: {meta.job_config.steps[0]["type"]=}'
        )

    # Load the source code into the UDF model
    udf_dict = meta.job_config.steps[0]["udf"]
    source_file_name = udf_dict["source"]

    code = load_callback(source_file_name).decode("utf-8")
    udf_dict["code"] = code
    del udf_dict["source"]

    # Do the same for headers
    for header_dict in udf_dict["headers"]:
        header_source_file_name = header_dict.get("source_file")
        if header_source_file_name:
            del header_dict["source_file"]
            header_code = load_callback(header_source_file_name).decode("utf-8")
            header_dict["source_code"] = header_code

    return GeoPandasUdfV2.model_validate(udf_dict)



from __future__ import annotations

import warnings
from datetime import datetime, timedelta
from typing import IO, TYPE_CHECKING, Any, List, Optional, Union

from pydantic import BaseModel, Field, StrictBool, StrictInt, StrictStr

from fused._formatter.formatter_jobs import fused_jobs_repr, fused_runresponse_repr
from fused._global_api import get_api
from fused.warnings import FusedInProgressWarning

if TYPE_CHECKING:
    from fused.api import FusedAPI
    from fused.models import JobConfig, JobStepConfig


class RunResponse(BaseModel):
    job_id: StrictStr
    """The identifier of this job."""
    job_name: StrictStr

    instance_id: Optional[StrictStr] = None
    instance_type: Optional[StrictStr] = None
    status: StrictStr
    """The status of the instance running this job."""
    terminal_status: StrictBool

    def _repr_html_(self) -> str:
        return fused_runresponse_repr(self)

    @property
    def _api(self) -> FusedAPI:
        # Note that this does not import the FusedAPI class for circular import reasons
        # We assume that the API has already been instantiated before a model is created
        return get_api()

    def get_status(self) -> RunResponse:
        """Fetch the status of this job

        Returns:
            The status of the given job.
        """
        return self._api.get_status(self)

    def get_logs(self, since_ms: Optional[int] = None) -> List[Any]:
        """Fetch logs for this job

        Args:
            since_ms: Timestamp, in milliseconds since epoch, to get logs for. Defaults to None for all logs.

        Returns:
            Log messages for the given job.
        """
        return self._api.get_logs(self, since_ms=since_ms)

    def print_logs(
        self, since_ms: Optional[int] = None, file: Optional[IO] = None
    ) -> None:
        """Fetch and print logs for this job

        Args:
            since_ms: Timestamp, in milliseconds since epoch, to get logs for. Defaults to None for all logs.
            file: Where to print logs to. Defaults to sys.stdout.

        Returns:
            None
        """
        logs = self.get_logs(since_ms=since_ms)
        for m in logs:
            print(m["message"].strip(), file=file)

    def get_exec_time(self) -> timedelta:
        """Determine the execution time of this job, using the logs.

        Returns:
            Time the job took. If the job is in progress, time from first to last log message is returned.
        """
        refreshed = self.get_status()
        if not refreshed.terminal_status:
            warnings.warn(
                FusedInProgressWarning(
                    f"Job is still in progress. Status: {refreshed.status}"
                )
            )

        logs = refreshed.get_logs()
        if len(logs):
            first_log = logs[0]
            last_log = logs[-1]

            first_ts = first_log["timestamp"]
            last_ts = last_log["timestamp"]

            d_ts = last_ts - first_ts
            return timedelta(milliseconds=d_ts)
        else:
            raise ValueError(
                "Job does not have logs, so execution time can't be determined"
            )

    def tail_logs(
        self,
        refresh_seconds: float = 1,
        sample_logs: bool = True,
        timeout: Optional[float] = None,
        get_logs_retries: int = 1,
    ) -> None:
        """Continuously print logs for this job

        Args:
            refresh_seconds: how frequently, in seconds, to check for new logs. Defaults to 1.
            sample_logs: if true, print out only a sample of logs. Defaults to True.
            timeout: if not None, how long to continue tailing logs for. Defaults to None for indefinite.
            get_logs_retries: Number of additional retries for log requests. Defaults to 1.
        """
        return self._api.tail_logs(
            self,
            refresh_seconds=refresh_seconds,
            sample_logs=sample_logs,
            timeout=timeout,
            get_logs_retries=get_logs_retries,
        )

    def wait_for_job(
        self,
        poll_interval_seconds: float = 5,
        timeout: Optional[float] = None,
    ) -> RunResponse:
        """Block the Python kernel until this job has finished

        Args:
            poll_interval_seconds: How often (in seconds) to poll for status updates. Defaults to 5.
            timeout: The length of time in seconds to wait for the job. Defaults to None.

        Raises:
            TimeoutError: if waiting for the job timed out.

        Returns:
            The status of the given job.
        """
        return self._api.wait_for_job(
            self,
            poll_interval_seconds=poll_interval_seconds,
            timeout=timeout,
        )

    def cancel(self) -> RunResponse:
        """Cancel this job

        Returns:
            A new job object.
        """
        return self._api.cancel_job(self)

    def refresh_status(self) -> RunResponse:
        """Refresh the status of this job."""
        if not self.terminal_status:
            new_status = self._api.get_status(self)
            self.job_id = new_status.job_id
            self.instance_id = new_status.instance_id
            self.instance_type = new_status.instance_type
            self.terminal_status = new_status.terminal_status
            self.status = new_status.status

        return self

    def __repr__(self) -> str:
        self.refresh_status()
        return super().__repr__()

    @classmethod
    def from_job_id(cls, job: Union[str, RunResponse]) -> RunResponse:
        """Creates a RunResponse object from either a job ID or a RunResponse.

        Args:
            job: Either a job ID string, or a RunResponse.

        Returns:
            A RunResponse object.
        """
        if isinstance(job, str):
            return RunResponse(
                job_id=job,
                job_name="null",
                instance_id=None,
                instance_type=None,
                status="null",
                terminal_status=True,
            )
        return job

    def get_results(self, arg_index: int = -1) -> Any | list[Any]:
        return self._api.get_job_results(self.job_id, arg_index)


class JobResponse(BaseModel):
    id: StrictStr
    """The identifier of this job."""
    user_id: StrictStr
    creation_date: datetime
    """When this job was started."""

    job_status: Optional[StrictStr] = None
    """The status of this job."""
    job_status_date: Optional[datetime] = None
    """When the job_status was last updated"""

    instance_id: Optional[StrictStr] = None
    instance_type: Optional[StrictStr] = None
    execution_environment_id: str

    @property
    def _api(self) -> FusedAPI:
        # Note that this does not import the FusedAPI class for circular import reasons
        # We assume that the API has already been instantiated before a model is created
        return get_api()

    def config(self) -> JobStepConfig:
        """Fetch the job step configuration

        Returns:
            The configuration for the job step in this job.
        """
        return self._api.get_job_config(input)._to_job_step_config()

    def steps_config(self) -> JobConfig:
        """Fetch the job configuration

        Returns:
            The configuration for all job steps in this job.
        """
        return self._api.get_job_config(input)

    def get_status(self) -> RunResponse:
        """Fetch the status of this job

        Returns:
            The status of the given job.
        """
        return self._api.get_status(self)

    def get_logs(self, since_ms: Optional[int] = None) -> List[Any]:
        """Fetch logs for this job

        Args:
            since_ms: Timestamp, in milliseconds since epoch, to get logs for. Defaults to None for all logs.

        Returns:
            Log messages for the given job.
        """
        return self._api.get_logs(self, since_ms=since_ms)

    def print_logs(
        self, since_ms: Optional[int] = None, file: Optional[IO] = None
    ) -> None:
        """Fetch and print logs for this job

        Args:
            since_ms: Timestamp, in milliseconds since epoch, to get logs for. Defaults to None for all logs.
            file: Where to print logs to. Defaults to sys.stdout.

        Returns:
            None
        """
        logs = self.get_logs(since_ms=since_ms)
        for m in logs:
            print(m["message"].strip(), file=file)

    def get_exec_time(self) -> timedelta:
        """Determine the execution time of this job, using the logs.

        Returns:
            Time the job took. If the job is in progress, time from first to last log message is returned.
        """
        refreshed = self.get_status()
        if not refreshed.terminal_status:
            warnings.warn(
                FusedInProgressWarning(
                    f"Job is still in progress. Status: {refreshed.status}"
                )
            )

        logs = refreshed.get_logs()
        if len(logs):
            first_log = logs[0]
            last_log = logs[-1]

            first_ts = first_log["timestamp"]
            last_ts = last_log["timestamp"]

            d_ts = last_ts - first_ts
            return timedelta(milliseconds=d_ts)
        else:
            raise ValueError(
                "Job does not have logs, so execution time can't be determined"
            )

    def tail_logs(
        self,
        refresh_seconds: float = 1,
        sample_logs: bool = True,
        timeout: Optional[float] = None,
        get_logs_retries: int = 1,
    ):
        """Continuously print logs for this job

        Args:
            refresh_seconds: how frequently, in seconds, to check for new logs. Defaults to 1.
            sample_logs: if true, print out only a sample of logs. Defaults to True.
            timeout: if not None, how long to continue tailing logs for. Defaults to None for indefinite.
            get_logs_retries: Number of additional retries for log requests. Defaults to 1.
        """
        return self._api.tail_logs(
            self,
            refresh_seconds=refresh_seconds,
            sample_logs=sample_logs,
            timeout=timeout,
            get_logs_retries=get_logs_retries,
        )

    def wait_for_job(
        self,
        poll_interval_seconds: float = 5,
        timeout: Optional[float] = None,
    ) -> RunResponse:
        """Block the Python kernel until this job has finished

        Args:
            poll_interval_seconds: How often (in seconds) to poll for status updates. Defaults to 5.
            timeout: The length of time in seconds to wait for the job. Defaults to None.

        Raises:
            TimeoutError: if waiting for the job timed out.

        Returns:
            The status of the given job.
        """
        return self._api.wait_for_job(
            self,
            poll_interval_seconds=poll_interval_seconds,
            timeout=timeout,
        )

    def cancel(self) -> RunResponse:
        """Cancel this job

        Returns:
            A new job object.
        """
        return self._api.cancel_job(self)


class Jobs(BaseModel):
    jobs: List[JobResponse]
    """The list of jobs."""

    # Options needed to refresh the same job list
    n: StrictInt = Field(..., repr=False)
    skip: StrictInt = Field(..., repr=False)
    per_request: StrictInt = Field(..., repr=False)
    max_requests: Optional[StrictInt] = Field(..., repr=False)

    @property
    def _api(self) -> FusedAPI:
        return get_api()

    def refresh(self) -> Jobs:
        """Returns this object with an updated job list"""
        new_jobs = self._api.get_jobs(
            n=self.n,
            skip=self.skip,
            per_request=self.per_request,
            max_requests=self.max_requests,
        )
        self.jobs = new_jobs.jobs
        return self

    def __repr__(self) -> str:
        self.refresh()
        txt = ""
        for job in self.jobs:
            txt += f"{job.id}: {job.job_status}\n"

        return txt

    def _repr_html_(self) -> str:
        return fused_jobs_repr(self)


CoerceableToJobId = Union[str, RunResponse, JobResponse]


def _object_to_job_id(job: CoerceableToJobId) -> str:
    if isinstance(job, RunResponse):
        return job.job_id
    elif isinstance(job, JobResponse):
        return job.id
    else:
        return job



from typing import Any, Optional

from fused_batch.api.api import FusedAPI
from fused_batch.models.api import UdfJobStepConfig
from fused_batch.models.udf import RootAnyBaseUdf
from fused_batch.models.udf._eval_result import UdfEvaluationResult
from fused_batch.models.udf.output import Output


def get_step_config_from_server(
    email: Optional[str],
    slug: str,
    cache_key: Any,
    _is_public: bool = False,
) -> UdfJobStepConfig:
    # cache_key is unused
    api = FusedAPI()
    if _is_public:
        obj = api._get_public_udf(slug)
    else:
        obj = api._get_udf(email, slug)
    udf = RootAnyBaseUdf.model_validate_json(obj["udf_body"]).root

    step_config = UdfJobStepConfig(udf=udf)
    return step_config


def get_github_udf_from_server(
    url: str,
    cache_key: Any,
):
    # cache_key is unused
    # TODO: Do this locally in fused-py
    api = FusedAPI(credentials_needed=False)
    obj = api._get_code_by_url(url)
    udf = RootAnyBaseUdf.model_validate_json(obj["udf_body"]).root

    step_config = UdfJobStepConfig(udf=udf)
    return step_config


def run_and_get_data(udf, *args, **kwargs):
    # TODO: This is a silly way to do this, because we have to pass parameters in such an odd way
    job = udf(*args, **kwargs)
    result = job.run_local()
    if isinstance(result, (Output, UdfEvaluationResult)):
        return result.data
    else:
        return result



# ruff: noqa: F401
from fused_batch._public_api import load_job
from fused_batch.models.api import UdfAccessToken, UdfJobStepConfig
from fused_batch.models.udf import AnyBaseUdf, BaseUdf, GeoPandasUdfV2
from fused_batch.models.udf.base_udf import AttrDict
from fused_batch.models.udf.common import Chunk, Chunks



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict

from fused._global_api import get_api

if TYPE_CHECKING:
    from fused.api import FusedAPI

# Pydantic methods to remove in __dir__
PYDANTIC_METHODS = {
    "Config",
    "construct",
    "copy",
    "from_orm",
    "json",
    "parse_file",
    "parse_obj",
    "schema",
    "schema_json",
    "update_forward_refs",
    "validate",
    "model_config",
    "model_construct",
    "model_copy",
    "model_dump",
    "model_dump_json",
    "model_extra",
    "model_fields",
    "model_fields_set",
    "model_json_schema",
    "model_parametrized_name",
    "model_post_init",
    "model_rebuild",
    "model_validate",
    "model_validate_json",
    "model_validate_strings",
    "parse_raw",
    "model_computed_fields",
}


class FusedBaseModel(BaseModel):
    @property
    def _api(self) -> FusedAPI:
        # Note that this does not import the FusedAPI class for circular import reasons
        # We assume that the API has already been instantiated before a model is created
        return get_api()

    def __dir__(self) -> List[str]:
        """Provide method name lookup and completion. Only provide 'public'
        methods.
        """
        # This enables autocompletion.
        normal_dir = {
            name
            for name in dir(type(self))
            if not name.startswith("_") and name not in PYDANTIC_METHODS
        }
        pydantic_fields = set(self.model_fields.keys())

        return sorted(normal_dir | pydantic_fields)

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)


UserMetadataType = Optional[Dict[str, Any]]



from typing import Callable, Dict, Optional, Sequence, Union

import geopandas as gpd

from fused_batch.models.api import Dataset


def create_href_mapping(
    stac_gdf: Union[Dataset, gpd.GeoDataFrame],
    signer: Callable[[str], str],
    asset_names: Optional[Sequence[str]] = None,
) -> Dict[str, str]:
    """Create an href mapping given a STAC GeoDataFrame and a lambda signer

    Args:
        stac_gdf: A GeoDataFrame representing STAC Items
        signer: A function applied on each asset href.
        asset_names: The keys in the assets dictionary to sign. Defaults to None, in which case it signs all assets.

    Returns:
        A dict mapping from original asset hrefs to signed asset hrefs.
    """
    if isinstance(stac_gdf, Dataset):
        stac_gdf = stac_gdf.get_dataframe(file_id=None, chunk_id=None)

    assets_series = stac_gdf["assets"]

    # Note: this assumes all assets have the same names
    all_asset_names = set(assets_series[0].keys())

    if asset_names is not None:
        required_asset_names = set(asset_names).intersection(all_asset_names)
    else:
        required_asset_names = all_asset_names

    href_mapping: Dict[str, str] = {}
    for assets in assets_series:
        for required_asset_name in required_asset_names:
            href = assets[required_asset_name]["href"]
            signed_href = signer(href)
            href_mapping[href] = signed_href

    return href_mapping



from __future__ import annotations

import json
import shlex
import subprocess
from base64 import b64encode
from functools import lru_cache
from io import SEEK_SET, BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Sequence, Union
from uuid import uuid4

from pydantic import BaseModel

from fused._global_api import set_api
from fused._optional_deps import HAS_PANDAS, PD_DATAFRAME
from fused._options import DEV_DEFAULT_BASE_URL, STAGING_DEFAULT_BASE_URL
from fused._options import options as OPTIONS
from fused.api.api import FusedAPI
from fused.models.api import JobConfig, JobStepConfig

if TYPE_CHECKING:
    import pandas as pd

# TODO: Consider making this us-west-2, depending on where the user is logged in
# DEFAULT_REPOSITORY = "926411091187.dkr.ecr.us-east-1.amazonaws.com/fused-job2"
DEFAULT_REPOSITORY = "926411091187.dkr.ecr.us-west-2.amazonaws.com/fused-job2"
DEFAULT_TAG = "latest"

MAX_ERROR_MESSAGE_SIZE = 10000
JOB_MOUNT_PATH = "/job/job.json"
JOB_INPUT_MOUNT_PATH = "/job/input"
DEFAULT_JOB_INPUT_HOST = "/tmp"


class DockerRunnable(BaseModel):
    command: str

    def run_and_get_bytes(self) -> bytes:
        """
        Run the command and return the bytes written to stdout.

        Raises an exception if the return code is not 0.
        """
        # TODO: Disable shell here
        # Check is false because we check it ourselves next
        p = subprocess.run(self.command, shell=True, capture_output=True, check=False)
        if p.returncode:
            error_message = p.stderr.decode("utf-8")
            if len(error_message) > MAX_ERROR_MESSAGE_SIZE:
                error_message = (
                    error_message[:MAX_ERROR_MESSAGE_SIZE] + "... (truncated)"
                )
            if len(error_message) == 0:
                error_message = "No message on stderr"
            raise ValueError(f"Failed ({p.returncode}): {error_message}")
        return p.stdout

    def run_and_get_output(self) -> str:
        """
        Run the command and return the utf-8 string written to stdout.

        Raises an exception if the return code is not 0.
        """
        return self.run_and_get_bytes().decode("utf-8")

    def run_and_tail_output(self) -> None:
        """
        Run the command and print output to stdout.

        Raises an exception if the return code is not 0.
        """
        # TODO: Disable shell here
        subprocess.run(self.command, shell=True, check=True)


class FusedDockerAPI(FusedAPI):
    """API for running jobs in the Fused Docker container."""

    repository: str
    tag: str
    mount_aws_credentials: bool
    mount_data_directory: Optional[str]
    mount_job_input_directory: Optional[str]
    additional_docker_args: Sequence[str]
    docker_command_wrapper: Optional[Callable[[str], str]]
    _auth_token: Optional[str]
    is_staging: bool
    is_gcp: bool
    is_aws: bool

    def __init__(
        self,
        *,
        repository: str = DEFAULT_REPOSITORY,
        tag: str = DEFAULT_TAG,
        mount_aws_credentials: bool = False,
        mount_data_directory: Optional[str] = None,
        mount_job_input_directory: Optional[str] = DEFAULT_JOB_INPUT_HOST,
        additional_docker_args: Sequence[str] = (),
        docker_command_wrapper: Optional[Callable[[str], str]] = None,
        auth_token: Optional[str] = None,
        auto_auth_token: bool = True,
        set_global_api: bool = True,
        is_staging: Optional[bool] = None,
        is_gcp: bool = False,
        is_aws: bool = False,
        pass_config_as_file: bool = True,
        base_url: Optional[str] = None,
        credentials_needed: bool = True,
    ):
        """Create a FusedDockerAPI instance.

        Keyword Args:
            repository: Repository name for jobs to start.
            tag: Tag name for jobs to start. Defaults to `'latest'`.
            mount_aws_credentials: Whether to add an additional volume for AWS credentials in the job. Defaults to False.
            mount_data_directory: If not None, path on the host to mount as the /data directory in the container. Defaults to None.
            mount_job_input_directory: If not None, path on the host to mount as the /job/input/ directory in the container. Defaults to None.
            additional_docker_args: Additional arguments to pass to Docker. Defaults to empty.
            docker_command_wrapper: Command to wrap the Docker execution in, e.g. `'echo {} 1>&2; exit 1'`. Defaults to None for no wrapping.
            auth_token: Auth token to pass to the Docker command. Defaults to automatically detect when auto_auth_token is True.
            auto_auth_token: Obtain the auth token from the (previous) global Fused API. Defaults to True.
            set_global_api: Set this as the global API object. Defaults to True.
            is_staging: Set this if connecting to the Fused staging environment. Defaults to None to automatically detect.
            is_gcp: Set this if running in GCP. Defaults to False.
            is_aws: Set this if running in AWS. Defaults to False.
            pass_config_as_file: If True, job configurations are first written to a temporary file and then passed to Docker. Defaults to True.
        """
        super().__init__(
            base_url=base_url,
            credentials_needed=credentials_needed,
            set_global_api=False,
        )
        self.repository = repository
        self.tag = tag
        self.mount_aws_credentials = mount_aws_credentials
        self.mount_data_directory = mount_data_directory
        self.mount_job_input_directory = mount_job_input_directory
        self.additional_docker_args = additional_docker_args
        self.docker_command_wrapper = docker_command_wrapper
        self.is_gcp = is_gcp
        self.is_aws = is_aws
        self.pass_config_as_file = pass_config_as_file
        if auth_token or not auto_auth_token:
            self._auth_token = auth_token
        else:
            self._auth_token = FusedAPI(set_global_api=False).auth_token()

        if is_staging is not None:
            self.is_staging = is_staging
        else:
            # Autodetect whether staging flag should be set based on base_url.
            self.is_staging = OPTIONS.base_url in (
                STAGING_DEFAULT_BASE_URL,
                DEV_DEFAULT_BASE_URL,
            )

        if set_global_api:
            set_api(self)

    def start_job(
        self,
        config: Union[JobConfig, JobStepConfig],
        *,
        additional_env: Optional[Sequence[str]] = ("FUSED_CREDENTIAL_PROVIDER=ec2",),
        **kwargs,
    ) -> DockerRunnable:
        """Execute an operation

        Args:
            config: the configuration object to run in the job.

        Keyword Args:
            additional_env: Any additional environment variables to be passed into the job, each in the form KEY=value. Defaults to None.
        """
        assert (
            kwargs.get("region") is None
        ), "region may not be specified with FusedDockerAPI"
        assert (
            kwargs.get("instance_type") is None
        ), "instance_type may not be specified with FusedDockerAPI"
        assert (
            kwargs.get("disk_size_gb") is None
        ), "disk_size_gb may not be specified with FusedDockerAPI"
        assert (
            kwargs.get("image_name") is None
        ), "image_name may not be specified with FusedDockerAPI"

        if isinstance(config, JobStepConfig):
            config = JobConfig(steps=[config])

        if self.pass_config_as_file:
            config_path = self._make_config_path(config.model_dump_json())
            args = ["--config-from-file", JOB_MOUNT_PATH]
        else:
            config_path = None
            args = ["--config", config.model_dump_json()]

        # TODO: This should return a RunResponse
        return self._make_run_command(
            "run-config", args, env=additional_env, config_path=config_path
        )

    def _make_run_command(
        self,
        command: str,
        args: Sequence[str],
        env: Optional[Sequence[str]] = None,
        config_path: Optional[str] = None,
    ) -> DockerRunnable:
        docker_args_part: List[str] = [*self.additional_docker_args]
        if env:
            for e in env:
                docker_args_part.append("-e")
                docker_args_part.append(shlex.quote(e))
        # TODO: Send in the base_url
        if self._auth_token:
            docker_args_part.append("-e")
            docker_args_part.append(f"FUSED_AUTH_TOKEN={self._auth_token}")
        if self.is_staging:
            docker_args_part.append("-e")
            docker_args_part.append("__FUSED_STAGING_LICENSE_CHECK=1")
        if self.mount_aws_credentials:
            docker_args_part.append("-v")
            docker_args_part.append(
                '"$HOME/.aws/credentials:/root/.aws/credentials:ro"'
            )
        if self.mount_data_directory is not None:
            docker_args_part.append("--mount")
            docker_args_part.append(
                f"type=bind,src={self.mount_data_directory},target=/data"
            )
            docker_args_part.append("-e")
            docker_args_part.append("FUSED_DATA_DIRECTORY=/data")
        if config_path is not None:
            docker_args_part.append("--mount")
            docker_args_part.append(
                f"type=bind,src={config_path},target={JOB_MOUNT_PATH}"
            )
        if self.mount_job_input_directory is not None:
            docker_args_part.append("--mount")
            docker_args_part.append(
                f"type=bind,src={self.mount_job_input_directory},target={JOB_INPUT_MOUNT_PATH}"
            )
        if self.is_gcp:
            docker_args_part.append("-e")
            docker_args_part.append("FUSED_GCP=1")
        if self.is_aws:
            docker_args_part.append("-e")
            docker_args_part.append("FUSED_AWS=1")

        args_part = [shlex.quote(a) for a in args]

        docker_image_name = f"{self.repository}:{self.tag}"

        job_command = shlex.quote(command)
        docker_command = f"docker run {' '.join(docker_args_part)} --rm {docker_image_name} {job_command} {' '.join(args_part)}"
        if self.docker_command_wrapper is not None:
            docker_command = self.docker_command_wrapper(docker_command)

        return DockerRunnable(command=docker_command)

    def _make_config_path(self, config_json: str) -> str:
        temp_file_name = f"/tmp/job_{uuid4()}.json"

        # TODO: Confirm temp_file_name doesn't already exist

        # We don't directly create the file in order to support running over SSH
        create_command = f"cat | base64 --decode > {temp_file_name}"
        if self.docker_command_wrapper is not None:
            create_command = self.docker_command_wrapper(create_command)

        b64data = b64encode(config_json.encode("utf-8"))

        # Don't capture output so that if the command errors, the output gets
        # sent back to the user
        subprocess.run(
            create_command,
            shell=True,
            check=True,
            input=b64data,
        )
        return temp_file_name

    def _replace_df_input(
        self, input: Union[str, List[str], "pd.DataFrame"]
    ) -> Union[str, List[str]]:
        replacement_bytes: Optional[bytes] = None
        extension: Optional[str] = None
        if HAS_PANDAS and isinstance(input, PD_DATAFRAME):
            with BytesIO() as tmp:
                input.to_parquet(tmp)
                tmp.seek(0, SEEK_SET)
                replacement_bytes = tmp.getvalue()
                extension = "parquet"
        elif isinstance(input, Path):
            with open(input, "rb") as f:
                replacement_bytes = f.read()
                extension = input.name.split(".", 1)[-1]

        if replacement_bytes is not None:
            assert (
                self.mount_job_input_directory
            ), "mount_job_input_directory must be set to pass DataFrame input"
            temp_file_name = f"df_{uuid4()}.{extension}"
            host_file_name = f"{Path(self.mount_job_input_directory) / temp_file_name}"
            container_file_name = f"file://{JOB_INPUT_MOUNT_PATH}/{temp_file_name}"

            # TODO: Confirm temp_file_name doesn't already exist

            # We don't directly create the file in order to support running over SSH
            # We use "base64 --decode" because Parquet as a binary file could interact with the shell in bad ways
            create_command = f"cat | base64 --decode > {host_file_name}"
            if self.docker_command_wrapper is not None:
                create_command = self.docker_command_wrapper(create_command)

            b64data = b64encode(replacement_bytes)

            # Don't capture output so that if the command errors, the output gets
            # sent back to the user
            subprocess.run(
                create_command,
                shell=True,
                check=True,
                input=b64data,
            )
            return container_file_name
        else:
            return input

    def _health(self) -> bool:
        """Check the health of the API backend"""
        runnable = self._make_run_command("version", [])
        runnable.run_and_get_output()
        return True

    @lru_cache
    def dependency_whitelist(self) -> Dict[str, str]:
        runnable = self._make_run_command("dependency-whitelist", [])
        content = runnable.run_and_get_bytes()
        return json.loads(content)


def ssh_command_wrapper(conn_string: str) -> Callable[[str], str]:
    """Creates a command wrapper that connects via SSH and sudo runs the command."""
    return (
        lambda command: f'ssh {conn_string} -t "sudo sh -c "{shlex.quote(shlex.quote(command))}""'
    )


def gcloud_command_wrapper(
    conn_string: str, *, zone: Optional[str] = None, project: Optional[str] = None
) -> Callable[[str], str]:
    """Creates a command wrapper that connects via gcloud and runs the command."""
    zone_arg = f"--zone {zone}" if zone else ""
    project_arg = f"--project {project}" if project else ""
    return (
        lambda command: f"gcloud compute ssh {zone_arg} {project_arg} {conn_string} --command {shlex.quote(command)}"
    )



# ruff: noqa: F401

from . import _cache, _magic, core, models, types, warnings
from ._load_udf import load
from ._options import PREVIEW_DEFAULT_BASE_URL as _PREVIEW_DEFAULT_BASE_URL
from ._options import PROD_DEFAULT_BASE_URL as _PROD_DEFAULT_BASE_URL
from ._options import STAGING_DEFAULT_BASE_URL as _STAGING_DEFAULT_BASE_URL
from ._options import _env, options, set_option
from ._public_api import _whoami, delete
from ._public_api import download as _download
from ._public_api import (
    get,
    get_udfs,
    ingest,
    ingest_nongeospatial,
    list,
    sign_url,
    sign_url_prefix,
    upload,
)
from ._run import run
from ._udf import udf
from ._utils import utils
from .api import FusedAPI, FusedDockerAPI, FusedDockerHTTPAPI, NotebookCredentials
from .core import cache, download, get_chunk_from_table, get_chunks_metadata

# Must be later so that models get imported first
from . import _experimental as experimental  # isort:skip

# Version is set to 0.0.0 as the version string is automatically generated by
# https://github.com/mtkennerly/poetry-dynamic-versioning
__version__ = "1.15.2"



from __future__ import annotations

from typing import Dict, Optional


def context_get_user_email() -> str:
    from fused_batch.api.api import FusedAPI

    api = FusedAPI()
    return api._whoami()["email"]


def context_get_auth_header(*, missing_ok: bool = False) -> Dict[str, str]:
    from fused_batch._auth import AUTHORIZATION

    if AUTHORIZATION.is_configured() or not missing_ok:
        return {"Authorization": f"Bearer {AUTHORIZATION.credentials.access_token}"}
    else:
        # Not logged in and that's OK
        return {}


def context_get_auth_scheme_and_token() -> Optional[tuple[str, str]]:
    return None


def context_in_realtime():
    return False



import ast
import json
import warnings
from enum import Enum
from functools import cached_property
from io import IOBase
from pathlib import PurePath
from textwrap import dedent, indent
from typing import Any, Dict, Iterable, Literal, Optional, Sequence, Tuple, Union
from uuid import UUID

import requests
from IPython import get_ipython
from IPython.core.inputsplitter import IPythonInputSplitter
from pydantic import BaseModel, ConfigDict, RootModel, field_validator

from fused_batch._formatter.udf import fused_header_repr, fused_udf_repr
from fused_batch._options import options as OPTIONS
from fused_batch.models.base import UserMetadataType
from fused_batch.models.udf.header import Header

from .._codegen import (
    extract_parameters,
    stringify_headers,
    stringify_input,
    stringify_named_params,
    structure_params,
)
from .._inplace import _maybe_inplace

METADATA_FUSED_ID = "fused:id"
METADATA_FUSED_SLUG = "fused:slug"
METADATA_FUSED_EXPLORER_TAB = "fused:explorerTab"


class HeaderSequence(RootModel[Sequence[Header]]):
    def _repr_html_(self) -> str:
        return fused_header_repr(self)

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]

    def __setitem__(self, item, value):
        self.root[item] = value

    def __len__(self):
        return len(self.root)


class AttrDict(dict):
    """Dictionary where keys can also be accessed as attributes"""

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            if __name in self:
                return self[__name]
            else:
                raise

    def __dir__(self) -> Iterable[str]:
        return self.keys()


class UdfType(str, Enum):
    GEOPANDAS_V2 = "geopandas_v2"


class BaseUdf(BaseModel):
    name: Optional[str] = None
    type: Literal[UdfType.GEOPANDAS_V2]
    code: str
    headers: Union[HeaderSequence, Sequence[Header], None] = None
    metadata: UserMetadataType = None

    @field_validator("headers", mode="before")
    @classmethod
    def _process_headers(cls, headers):
        processed_headers = []
        if headers is not None:
            for header in headers:
                if isinstance(header, str):
                    module_name = PurePath(header).name.split(".", maxsplit=1)[0]
                    processed_headers.append(
                        Header.from_code(module_name=module_name, source_file=header)
                    )
                elif (
                    isinstance(header, dict)
                    and header.get("source_file")
                    and not header.get("source_code")
                ):  # headers = {'source_file': 'header.py', 'module_name': 'header'}
                    source_file = header["source_file"]
                    module_name = header.get(
                        "module_name",
                        PurePath(source_file).name.split(".", maxsplit=1)[0],
                    )
                    processed_headers.append(
                        Header.from_code(
                            module_name=module_name,
                            source_file=source_file,
                        )
                    )
                else:
                    processed_headers.append(header)

        return processed_headers

    @field_validator("code", mode="before")
    @classmethod
    def udf_is_str(cls, v):
        if isinstance(v, IOBase):
            # Handle passing in a file as the code
            v.seek(0)
            data = v.read()
            if isinstance(data, str):
                return data
            elif isinstance(data, bytes):
                return data.decode("utf-8")
            else:
                raise ValueError("Expected string or bytes from the file")

        return v

    @classmethod
    def from_gist(cls, gist_id: str):
        """Create a Udf from a GitHub gist."""
        # TODO: if a versioned gist, taking the last / won't work as that will be the
        # version, not the gist id.
        if "/" in gist_id:
            gist_id = gist_id.split("/")[-1]

        url = f"https://api.github.com/gists/{gist_id}"
        r = requests.get(url, timeout=OPTIONS.request_timeout)
        r.raise_for_status()

        files: dict = r.json()["files"]

        obj = {}
        # TODO: Update expected files schema
        assert "code.py" in files.keys()
        assert "parameters.json" in files.keys()

        obj["code"] = _fetch_gist_content(files, "code.py")

        if "table_schema.json" in files.keys():
            table_schema_text = _fetch_gist_content(files, "parameters.json")
            obj["table_schema"] = json.loads(table_schema_text)

        parameters_text = _fetch_gist_content(files, "parameters.json")
        parameters = json.loads(parameters_text)
        obj.update(parameters)

        return cls.model_validate(obj)

    def to_fused(
        self,
        slug: Optional[str] = ...,
        over_id: Union[str, UUID, None] = None,
        as_new: Optional[bool] = None,
        inplace: bool = True,
    ):
        """
        Save this UDF on the Fused service.

        Args:
            slug: ID to refer to this UDF as in URLs.
            over_id: ID to save the UDF over.
            as_new: If True, force saving this UDF as new.
            inplace: If True (default), update the UDF object with the new saved ID.
        """
        from fused_batch._global_api import get_api

        backend_id = (
            str(over_id) if over_id else self._get_metadata_safe(METADATA_FUSED_ID)
        )
        if backend_id is None or as_new:
            assert (
                as_new is not False
            ), "Cannot detect ID to save over, so as_new cannot be False."
            backend_id = None

        if slug is Ellipsis:
            slug = self._get_metadata_safe(METADATA_FUSED_SLUG)

        api = get_api()
        result = api.save_udf(
            udf=self,
            slug=slug,
            id=backend_id,
        )
        new_id = result["id"]

        ret = _maybe_inplace(self, inplace)
        ret._set_metadata_safe(METADATA_FUSED_ID, new_id)
        return ret

    def delete_saved(self, inplace: bool = True):
        from fused_batch._global_api import get_api

        backend_id = self._get_metadata_safe(METADATA_FUSED_ID)
        assert backend_id is not None, "No saved UDF ID found in metadata."

        api = get_api()
        api.delete_saved_udf(
            id=backend_id,
        )

        ret = _maybe_inplace(self, inplace)
        # ret.metadata must be non None because we read the backend ID from there
        ret.metadata.pop(METADATA_FUSED_ID)
        return ret

    def _repr_html_(self) -> str:
        return fused_udf_repr(self)

    def _get_metadata_safe(
        self, key: str, default: Optional[Any] = None
    ) -> Optional[str]:
        if self.metadata is not None:
            return self.metadata.get(key, default)
        else:
            return None

    def _set_metadata_safe(self, key: str, value: Any):
        if self.metadata is None:
            self.metadata = {}

        self.metadata[key] = value

    def _generate_code(
        self, include_imports=True, headerfile=False
    ) -> Tuple[str, Sequence[str]]:
        def _extract_fn(src: str) -> Union[ast.FunctionDef, ast.AsyncFunctionDef]:
            # TODO: handle Header objects in header param
            parsed_ast = ast.parse(src)

            # Find all function definitions in the AST
            function_defs = [
                node
                for node in ast.walk(parsed_ast)
                if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))
            ]

            # first function (assume it's the target function)
            return function_defs[0]

        def _extract_fn_body(
            function_def: Union[ast.FunctionDef, ast.AsyncFunctionDef], src: str
        ) -> str:
            target_function_body = function_def.body
            # Reconstruct the source code of the function body
            line_start = target_function_body[0].lineno - 1
            line_end = target_function_body[-1].end_lineno
            target_function_body_lines = [
                line for line in src.splitlines()[line_start:line_end]
            ]
            target_function_body_str = "\n".join(target_function_body_lines)

            return target_function_body_str

        # Derive parameters
        positional_parameters, named_parameters = extract_parameters(self.code)

        _params_fn_original = positional_parameters + stringify_named_params(
            named_parameters
        )
        # String: Imports
        str_imports = (
            "\n".join(
                [
                    "import fused_batch",
                    "from fused_batch.models.udf import Header",
                ]
            )
            + "\n\n"
        )
        if self.type == "map":
            # TODO: Perhaps use positional_parameters.insert(0, f'"{stringify_input(self.input)}"') instead
            positional_parameters[0] = f'"{stringify_input(self.input)}"'
        elif self.type in ["join", "join_singlefile"]:
            positional_parameters[0] = f'"{stringify_input(self.input_left)}"'
            positional_parameters[1] = f'"{stringify_input(self.input_right)}"'
        _params_fn = positional_parameters + stringify_named_params(
            named_parameters
        )  # Includes all named parameters
        # String: UDF header - Replace Header with file reference
        header_files = []
        processed_headers = []
        # TODO: do this conversion as part of header attribute
        for header in self.headers:
            if isinstance(header, Header):
                if headerfile:
                    filename = header.module_name + ".py"
                    processed_headers.append(filename)

                else:
                    processed_headers.append(header)
                # Create a file magic string
                header_files.append(header._generate_cell_code())

            else:
                processed_headers.append(header)

        _headers = (
            stringify_headers(processed_headers) if headerfile else processed_headers
        )
        _schema = repr(self.table_schema.to_string()) if self.table_schema else '""'
        params_decorator = [f"\n    headers={_headers}", f"\n    schema={_schema}\n"]
        str_udf_header = (
            f"@fused_batch.udf({structure_params(params_decorator, separator=',')})"
        )
        # String: Function header
        fn = _extract_fn(self.code)
        str_async = "async " if isinstance(fn, ast.AsyncFunctionDef) else ""
        str_fn_header = f"{str_async}def {self.entrypoint}({structure_params(_params_fn_original)}):"
        # String: Function body
        str_fn_body = dedent(_extract_fn_body(fn, src=self.code))

        str_udf = f"""
{str_udf_header}
{str_fn_header}
{indent(str_fn_body,  " " * 4)}
"""
        if include_imports:
            str_udf = str_imports + str_udf
        return str_udf.strip("\n"), header_files

    def render(self):
        # Get the current IPython instance.
        ipython = get_ipython()
        if ipython is None:
            raise RuntimeError("This function can only be used in a Jupyter Notebook.")

        # Create an instance of IPythonInputSplitter.
        splitter = IPythonInputSplitter()

        # Generate code string and split into lines.
        lines = self._generate_code()[0].strip().split("\n")

        # Set the content of the subsequent cell with.
        ipython.set_next_input(splitter.transform_cell("\n".join(lines)))

    @property
    def utils(self):
        return self._cached_utils

    @cached_property
    def _cached_utils(self):
        if len(self.headers) == 0:
            raise ValueError("UDF does not have a utils module")
        if len(self.headers) > 1:
            raise ValueError("UDF has multiple header modules")
        if self.headers[0].module_name != "utils":
            warnings.warn(
                f"Accessing header module {self.headers[0].module_name} under the name utils"
            )
        # TODO: Even though this might have already been evaluated, we have to evaluate it again now
        # It is at least cached

        vals = self.headers[0]._exec()

        return AttrDict(vals)

    model_config = ConfigDict(exclude={"utils"})


def _fetch_gist_content(gist_files_dict: Dict, fname: str) -> str:
    gist_data = gist_files_dict[fname]

    if gist_data["truncated"]:
        full_url = gist_data["raw_url"]
        full_r = requests.get(full_url, timeout=OPTIONS.request_timeout)
        full_r.raise_for_status()
        return full_r.text
    else:
        return gist_data["content"]



from typing import Any, List, Optional, Union

from pydantic import BaseModel, ConfigDict

from fused._formatter.formatter_eval_result import fused_eval_result_repr
from fused.models.schema import Schema
from fused.models.udf import AnyBaseUdf


class UdfEvaluationResult(BaseModel):
    data: Any = None
    sidecar: Optional[bytes] = None

    udf: Optional[AnyBaseUdf] = None
    table_schema: Optional[Schema] = None

    time_taken_seconds: float

    stdout: Optional[str] = None
    stderr: Optional[str] = None
    has_exception: bool = False
    exception_class: Optional[str] = None
    error_message: Optional[str] = None
    error_lineno: Optional[int] = None

    def _repr_html_(self) -> str:
        return fused_eval_result_repr(self)

    model_config = ConfigDict(arbitrary_types_allowed=True)


class MultiUdfEvaluationResult(BaseModel):
    udf_results: List[Union[UdfEvaluationResult, Any]]

    def _repr_html_(self) -> str:
        # Aggregate reprs
        result_reprs = [
            udf_result._repr_html_()
            if hasattr(udf_result, "_repr_html_")
            else repr(udf_result)
            for udf_result in self.udf_results
        ]
        return "<br><br>".join(result_reprs)

    model_config = ConfigDict(arbitrary_types_allowed=True)



import click

from fused_batch._auth import authenticate as _authenticate
from fused_batch.api import logout as _logout


@click.group()
def main():
    pass


@main.command()
def authenticate():
    _authenticate()


@main.command()
def logout():
    _logout()


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import shutil
from functools import lru_cache
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Union,
    overload,
)

import geopandas as gpd
import pandas as pd
import requests
from pydantic import BaseModel

from fused_batch._deserialize import (
    zip_to_join_args,
    zip_to_join_singlefile_args,
    zip_to_map_args,
)
from fused_batch._deserialize_parquet import parquet_to_df
from fused_batch._global_api import set_api
from fused_batch._options import options as OPTIONS
from fused_batch._request import raise_for_status
from fused_batch.api._open_dataset import post_open_table
from fused_batch.models import JoinInput, JoinSingleFileInput, MapInput
from fused_batch.models.api import (
    JobConfig,
    JobStepConfig,
    JoinJobStepConfig,
    JoinSinglefileJobStepConfig,
    ListDetails,
    MapJobStepConfig,
    Table,
)
from fused_batch.models.api._folder import Folder

DEFAULT_ENDPOINT = "http://localhost:8789"


class DockerHTTPRunnable(BaseModel):
    endpoint: str
    command: Any = None

    def run_and_get_bytes(self) -> bytes:
        """
        Run the command and return the bytes written to stdout.

        Raises an exception if an HTTP error status is returned.
        """
        r = requests.post(
            url=self.endpoint,
            json=self.command,
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.content

    def run_and_get_output(self) -> str:
        """
        Run the command and return the utf-8 string written to stdout.

        Raises an exception if an HTTP error status is returned.
        """
        return self.run_and_get_bytes().decode("utf-8")

    def run_and_tail_output(self) -> None:
        """
        Run the command and print output to stdout.

        Raises an exception if an HTTP error status is returned.
        """
        with requests.post(
            url=self.endpoint,
            json=self.command,
            timeout=OPTIONS.request_timeout,
            stream=True,
        ) as r:
            if r.encoding is None:
                r.encoding = "utf-8"
            raise_for_status(r)
            for line in r.iter_lines(decode_unicode=True):
                print(line)


class FusedDockerHTTPAPI:
    """API for running jobs in the Fused Docker container over an HTTP connection."""

    endpoint: str

    def __init__(
        self,
        endpoint: str = DEFAULT_ENDPOINT,
        *,
        set_global_api: bool = True,
    ):
        """Create a FusedDockerHTTPAPI instance.

        Args:
            endpoint: The HTTP endpoint to connect to. Defaults to `"http://localhost:8789"`.

        Keyword Args:
            set_global_api: Set this as the global API object. Defaults to True.
        """
        self.endpoint = endpoint.rstrip("/")

        if set_global_api:
            set_api(self)

    def sample_map(
        self,
        config: MapJobStepConfig,
        *,
        file_id: Optional[Union[str, int]] = None,
        chunk_id: Optional[int] = None,
        n_rows: Optional[int] = None,
    ) -> MapInput:
        """Fetch a sample of an operation

        Args:
            config: The configuration to sample from.

        Keyword Args:
            file_id: The identifier of this file. Defaults to None.
            chunk_id: The numeric index of the chunk within the file to fetch. Defaults to None.
            n_rows: The maximum number of rows to sample. Defaults to None for all rows in the chunk.
        """
        assert isinstance(config, MapJobStepConfig)
        sample_request = self._sample(
            config,
            file_id=file_id,
            chunk_id=chunk_id,
        )
        content = sample_request.run_and_get_bytes()
        return zip_to_map_args(content, n_rows=n_rows)

    def _whole_file_sample_map(
        self,
        config: MapJobStepConfig,
        *,
        file_id: Union[str, int],
        n_rows: Optional[int] = None,
    ) -> MapInput:
        assert isinstance(config, MapJobStepConfig)
        sample_request = self._sample(
            config,
            file_id=file_id,
            whole_file=True,
        )
        content = sample_request.run_and_get_bytes()
        return zip_to_map_args(content, n_rows=n_rows)

    def download_table_bbox(
        self,
        path: str,
        minx: float,
        miny: float,
        maxx: float,
        maxy: float,
        n_rows: Optional[int] = None,
        columns: Optional[List[str]] = None,
        clip: bool = True,
        buffer: Optional[float] = None,
    ) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
        args = [
            "bbox",
            "--output-to-stdout",
            "--base-path",
            path,
            "--bbox-minx",
            str(minx),
            "--bbox-miny",
            str(miny),
            "--bbox-maxx",
            str(maxx),
            "--bbox-maxy",
            str(maxy),
        ]
        if n_rows is not None:
            args.append("--n-rows")
            args.append(str(n_rows))
        if columns is not None:
            for column in columns:
                args.append("--columns")
                args.append(column)
        args.append("--clip" if clip else "--no-clip")
        if buffer is not None:
            args.append("--buffer")
            args.append(str(buffer))

        request = self._make_run_command("download-table", args)
        content = request.run_and_get_bytes()
        return parquet_to_df(content)

    def sample_join(
        self,
        config: JoinJobStepConfig,
        *,
        file_id: Optional[Union[str, int]] = None,
        chunk_id: Optional[int] = None,
        n_rows: Optional[int] = None,
    ) -> JoinInput:
        """Fetch a sample of an operation

        Args:
            config: The configuration to sample from.

        Keyword Args:
            file_id: The identifier of this file. Defaults to None.
            chunk_id: The numeric index of the chunk within the file to fetch. Defaults to None.
            n_rows: The maximum number of rows to sample from the left dataset. Defaults to None for all rows in the chunk.
        """
        assert isinstance(config, JoinJobStepConfig)
        sample_request = self._sample(
            config,
            file_id=file_id,
            chunk_id=chunk_id,
        )
        content = sample_request.run_and_get_bytes()
        return zip_to_join_args(content, n_rows=n_rows)

    def sample_single_file_join(
        self,
        config: JoinSinglefileJobStepConfig,
        *,
        file_id: Optional[Union[str, int]] = None,
        chunk_id: Optional[int] = None,
        n_rows: Optional[int] = None,
    ) -> JoinSingleFileInput:
        """Fetch a sample of an operation

        Args:
            config: The configuration to sample from.

        Keyword Args:
            file_id: The identifier of this file. Defaults to None.
            chunk_id: The numeric index of the chunk within the file to fetch. Defaults to None.
            n_rows: The maximum number of rows to sample from the left dataset. Defaults to None for all rows in the chunk.
        """
        assert isinstance(config, JoinSinglefileJobStepConfig)
        sample_request = self._sample(
            config,
            file_id=file_id,
            chunk_id=chunk_id,
        )
        content = sample_request.run_and_get_bytes()
        return zip_to_join_singlefile_args(content, n_rows=n_rows)

    def _sample(
        self,
        config: Union[MapJobStepConfig, JoinJobStepConfig, JoinSinglefileJobStepConfig],
        *,
        file_id: Optional[Union[str, int]] = None,
        chunk_id: Optional[int] = None,
        whole_file: bool = False,
    ) -> DockerHTTPRunnable:
        """Fetch a sample of an operation

        Args:
            config: The configuration to sample from.

        Keyword Args:
            file_id: The identifier of this file. Defaults to None.
            chunk_id: The numeric index of the chunk within the file to fetch. Defaults to None.
        """
        assert (
            isinstance(config, MapJobStepConfig)
            or isinstance(config, JoinJobStepConfig)
            or isinstance(config, JoinSinglefileJobStepConfig)
        )

        args = ["--output-to-stdout", "--config", config.model_dump_json()]

        if file_id is not None:
            args.append("--file-id")
            args.append(str(file_id))

        if chunk_id is not None:
            args.append("--chunk-id")
            args.append(str(chunk_id))

        if whole_file:
            args.append("--whole-file")

        return self._make_run_command("make-sample", args)

    def start_job(
        self,
        config: Union[JobConfig, JobStepConfig],
        **kwargs,
    ) -> DockerHTTPRunnable:
        """Execute an operation

        Args:
            config: the configuration object to run in the job.
        """
        assert (
            kwargs.get("region") is None
        ), "region may not be specified with FusedDockerHTTPAPI"
        assert (
            kwargs.get("instance_type") is None
        ), "instance_type may not be specified with FusedDockerHTTPAPI"
        assert (
            kwargs.get("disk_size_gb") is None
        ), "disk_size_gb may not be specified with FusedDockerHTTPAPI"
        assert (
            kwargs.get("additional_env") is None
        ), "additional_env may not be specified with FusedDockerHTTPAPI"
        assert (
            kwargs.get("image_name") is None
        ), "image_name may not be specified with FusedDockerHTTPAPI"

        if isinstance(config, JobStepConfig):
            config = JobConfig(steps=[config])

        args = ["--config", config.model_dump_json()]

        # TODO: This should return a RunResponse
        return self._make_run_command("run-config", args)

    def open_table(
        self,
        path: str,
        *,
        fetch_samples: Optional[bool] = None,
    ) -> Table:
        """Open a Table object given a path to the root of the table

        Args:
            path: The path to the root of the table on remote storage

        Keyword Args:
            fetch_samples: If True, fetch sample of each table when opening the dataset.

        Example:

            ```python
            table = fused.open_table(path="s3://my_bucket/path/to/dataset/table/")
            ```

        Returns:
            A Table object
        """
        args = ["--base-path", path, "--output-to-stdout"]

        runnable = self._make_run_command("open-table", args)
        content = runnable.run_and_get_bytes()
        table = Table.model_validate_json(content)
        post_open_table(table=table, fetch_samples=fetch_samples)
        return table

    def open_folder(
        self,
        path: str,
        *,
        fetch_minimal_table_metadata: Optional[bool] = None,
        fetch_table_metadata: Optional[bool] = None,
        fetch_samples: Optional[bool] = None,
        table_mode: bool = True,
        max_depth: Optional[int] = None,
    ) -> Folder:
        """Open all Table objects under the path.

        Args:
            path: The path to the root of the folder on remote storage

        Keyword Args:
            fetch_table_metadata: If True, fetch metadata on each table when getting dataset metadata.
            fetch_samples: If True, fetch sample of each table when opening the dataset.
            max_depth: Maximum depth of Tables to open. Beyond this tables will be opened virtually.

        Example:

            datasets = fused.open_folder(
                path="s3://my_bucket/path/to/folder/"
            )

        Returns:
            A list of Dataset objects
        """
        should_fetch_table_metadata = False
        if fetch_table_metadata is not None:
            should_fetch_table_metadata = fetch_table_metadata
        elif OPTIONS.open.fetch_table_metadata is not None:
            should_fetch_table_metadata = OPTIONS.open.fetch_table_metadata

        should_fetch_minimal_table_metadata = True
        if fetch_minimal_table_metadata is not None:
            should_fetch_minimal_table_metadata = fetch_minimal_table_metadata
        elif OPTIONS.open.fetch_minimal_table_metadata is not None:
            should_fetch_minimal_table_metadata = (
                OPTIONS.open.fetch_minimal_table_metadata
            )

        args = ["--base-path", path, "--output-to-stdout"]

        args.append("--fetch-table-metadata")
        args.append("True" if should_fetch_table_metadata else "False")
        args.append("--fetch-minimal-table-metadata")
        args.append("True" if should_fetch_minimal_table_metadata else "False")
        if max_depth is not None:
            args.append("--max-depth")
            args.append(f"{max_depth}")

        cmd = "open-table-folder" if table_mode else "open-dataset-folder"
        runnable = self._make_run_command(cmd, args)
        content = runnable.run_and_get_bytes()
        folder = Folder.model_validate_json(content)
        for table in folder.tables:
            post_open_table(
                table=table,
                fetch_samples=fetch_samples,
            )
        return folder

    def _make_run_command(
        self,
        command: str,
        args: Sequence[str],
    ) -> DockerHTTPRunnable:
        args_part = [command, *list(args)]
        return DockerHTTPRunnable(endpoint=f"{self.endpoint}/cli", command=args_part)

    def upload(self, path: str, data: Union[bytes, BinaryIO]) -> None:
        """Upload a binary blob to a cloud location"""
        raise NotImplementedError("upload is not implemented on FusedDockerAPI")

    def _replace_df_input(
        self, input: Union[str, List[str], gpd.GeoDataFrame]
    ) -> Union[str, List[str]]:
        if isinstance(input, gpd.GeoDataFrame):
            # TODO: Consider encoding in line?
            raise NotImplementedError(
                "Cannot pass GeoDataFrame input through FusedDockerHTTPAPI"
            )
        elif isinstance(input, Path):
            # TODO: Consider encoding in line?
            raise NotImplementedError(
                "Cannot pass Path input through FusedDockerHTTPAPI"
            )
        else:
            return input

    def _health(self) -> bool:
        """Check the health of the API backend"""
        r = requests.get(
            url=f"{self.endpoint}/health",
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return True

    @overload
    def list(self, path: str, *, details: Literal[True]) -> List[ListDetails]:
        ...

    @overload
    def list(self, path: str, *, details: Literal[False] = False) -> List[str]:
        ...

    def list(self, path: str, *, details: bool = False):
        args = ["--path", path]

        if details:
            args.append("--details")

        runnable = self._make_run_command("files-list", args)
        content = runnable.run_and_get_bytes()
        result = json.loads(content)
        if details:
            result = [ListDetails.model_validate(detail) for detail in result]
        return result

    def delete(self, path: str, max_deletion_depth: int | Literal["unlimited"]) -> bool:
        args = ["--path", path, "--max-deletion-depth", str(max_deletion_depth)]
        runnable = self._make_run_command("files-delete", args)
        content = runnable.run_and_get_bytes()
        return json.loads(content)

    def get(self, path: str) -> bytes:
        url = self.sign_url(path)
        r = requests.get(
            url=url,
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.content

    def download(self, path: str, local_path: Union[str, Path]) -> None:
        url = self.sign_url(path)
        r = requests.get(
            url=url,
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        with open(local_path, "wb") as f:
            shutil.copyfileobj(r.raw, f)

    def sign_url(self, path: str) -> str:
        args = ["--path", path]

        runnable = self._make_run_command("files-sign-url", args)
        content = runnable.run_and_get_bytes()
        return json.loads(content)

    def sign_url_prefix(self, path: str) -> Dict[str, str]:
        args = ["--path", path]

        runnable = self._make_run_command("files-sign-url-prefix", args)
        content = runnable.run_and_get_bytes()
        return json.loads(content)

    def _get_table_names(self, path: str) -> List[str]:
        tables: List[str] = []
        for table_path in self.list(path):
            table_name = table_path.split("/")[-1]
            tables.append(table_name)

        return tables

    @lru_cache
    def dependency_whitelist(self) -> Dict[str, str]:
        runnable = self._make_run_command("dependency-whitelist", [])
        content = runnable.run_and_get_bytes()
        return json.loads(content)



def add_migration(migration_func, cls, from_version):
    try:
        assert cls.__dict__["_validate_version"]
    except KeyError:
        raise ValueError(
            "Cannot add migration to un-versioned object. To version, add a '_validate_version: bool = True' attribute."
        )

    # Add migration to class
    _vobj__migrations = cls._vobj__migrations.copy()
    _vobj__migrations.append((from_version, migration_func, cls))

    # Sort by migration version
    cls._vobj__migrations = sorted(_vobj__migrations)


def migration(cls, from_version):
    """
    Decorator for adding a migration function to an object class. Use this
    decorator on any function or method that should be used for migrating an
    object from one version to another. This is an equivalent alternative to the
    versionedobject.object.add_migration function.

    :param cls: Class object to add migration to
    :param from_version: Version to migrate from. If you are migrating an object that\
        previously had no version number, use 'None' here.
    :param to_version: Version to migrate to
    """

    def _inner_migration(migration_func):
        add_migration(migration_func, cls, from_version)

    return _inner_migration


# @migration(JobStepConfig, "0.0.0")
# def passthrough_sample_migration(values: Dict[str, Any]) -> Dict[str, Any]:
#     return values



"""Vendored from xarray
https://github.com/pydata/xarray/blob/f4e0523be0ce8babaa8eff38365e5308b1fdb76b/xarray/core/formatting_html.py
"""

from __future__ import annotations

from html import escape
from typing import TYPE_CHECKING, List

from pydantic import BaseModel

from fused._formatter.common import copyable_text, load_static_files

if TYPE_CHECKING:
    from fused._options import Options


def _obj_repr(obj, header_components, sections):
    """Return HTML repr of an xarray object.

    If CSS is not injected (untrusted notebook), fallback to the plain text repr.

    """
    header = f"<div class='xr-header'>{''.join(h for h in header_components)}</div>"
    sections = "".join(f"<li class='xr-section-item'>{s}</li>" for s in sections)

    icons_svg, css_style = load_static_files()
    return (
        "<div>"
        f"{icons_svg}<style>{css_style}</style>"
        f"<pre class='xr-text-repr-fallback'>{escape(repr(obj))}</pre>"
        "<div class='xr-wrap' style='display:none'>"
        f"{header}"
        f"<ul class='xr-sections'>{sections}</ul>"
        "</div>"
        "</div>"
    )


def _copyable_toml(options: Options) -> str:
    return f'<div style="grid-column: 1 / span 4;">Copy as TOML {copyable_text(text=options._to_toml(), show_text=False)}</div>'


def _options_repr(obj: BaseModel) -> str:
    ret: List[str] = []
    for key in obj.model_dump().keys():
        value = getattr(obj, key)
        if isinstance(value, BaseModel):
            ret.append(f"{key}: {_options_repr(value)}")
        else:
            ret.append(f"{key}: <code>{escape(repr(value))}</code>")
    return f'<div style="grid-column: 1 / span 4;"><ul>{"".join([f"<li>{r}</li>" for r in ret])}</ul></div>'


def fused_options_repr(options: Options) -> str:
    header_components = ["<div class='xr-obj-type'>fused.options</div>"]

    sections: List[str] = []
    if hasattr(options, "_to_toml"):
        sections.append(_copyable_toml(options))
    sections.append(_options_repr(options))

    return _obj_repr(options, header_components, sections)



from typing import Optional, Sequence

from pydantic import BaseModel

from .dataset import Table


class Folder(BaseModel):
    base_path: Optional[str] = None
    tables: Sequence[Table]
    folders: Sequence[str] = ()



from __future__ import annotations

from typing import Any, Optional

from ._impl._context_impl import context_get_user_email
from ._impl._reimports import AnyBaseUdf
from ._impl._udf_ops_impl import get_github_udf_from_server, get_step_config_from_server


def load_udf_from_fused(
    email_or_id: str, id: Optional[str] = None, *, cache_key: Any = None
) -> AnyBaseUdf:
    """
    Download the code of a UDF, to be run inline.

    Args:
        email_or_id: Email of the UDF's owner, or name of the UDF to import.
        id: Name of the UDF to import. If only the first argument is provided, the current user's email will be used.

    Keyword args:
        cache_key: Additional cache key for busting the UDF cache
    """
    if id is None:
        id = email_or_id
        try:
            email = context_get_user_email()
        except Exception as e:
            raise ValueError(
                "could not detect user ID from context, please specify the UDF as 'user@example.com', 'udf_name'."
            ) from e
    else:
        email = email_or_id
    step_config = get_step_config_from_server(
        email=email,
        slug=id,
        cache_key=cache_key,
    )

    return step_config.udf


def load_udf_from_github(url: str, *, cache_key: Any = None) -> AnyBaseUdf:
    """
    Download the code of a UDF, to be run inline.

    Args:
        email_or_id: Email of the UDF's owner, or name of the UDF to import.
        id: Name of the UDF to import. If only the first argument is provided, the current user's email will be used.

    Keyword args:
        cache_key: Additional cache key for busting the UDF cache
    """
    step_config = get_github_udf_from_server(url=url)

    return step_config.udf



import json
from pathlib import Path
from typing import Any, Dict, List, Optional, TextIO, Tuple, Type, Union

import requests
from jinja2 import (
    BaseLoader,
    ChoiceLoader,
    DictLoader,
    Environment,
    FileSystemLoader,
    PackageLoader,
    StrictUndefined,
    Template,
    TemplateNotFound,
    select_autoescape,
)
from pydantic import BaseModel

from fused_batch._formatter.formatter_templates import (
    fused_template_repr,
    fused_templates_repr,
)
from fused_batch._options import options as OPTIONS
from fused_batch._request import raise_for_status
from fused_batch.models import (
    AnyJobStepConfig,
    JobConfig,
    JobStepConfig,
    RootAnyJobStepConfig,
)
from fused_batch.models.coerce_dataset import (
    CoerceableToDatasetInput,
    CoerceableToDatasetOutput,
    _object_to_dataset_input,
    _object_to_dataset_output,
)
from fused_batch.models.udf import AnyBaseUdf, RootAnyBaseUdf

from .loaders import RegisteredLoader, UrlLoader


class FusedTemplate:
    name: str
    """Name of this template."""
    source: str
    """The Jinja template source for the JSON representation of the object."""
    template: Template
    """The Jinja template."""
    defaults: Dict[str, Any]
    """Default keyword arguments for the template."""

    def __init__(
        self, name: str, source: str, template: Template, defaults: Dict[str, Any]
    ):
        self.name = name
        self.source = source
        self.template = template
        self.defaults = defaults

    def _repr_html_(self):
        return fused_template_repr(self)

    def print(self, **kwargs):
        """Print the code generated by the template.

        Keyword arguments are passed to the template."""
        raise NotImplementedError()

    def code(self, **kwargs):
        """Set the code generated by the template as the next IPython cell.

        Keyword arguments are passed to the template."""
        raise NotImplementedError()

    def render(self, **kwargs):
        """Render the template to an object.

        Keyword arguments are passed to the template."""
        raise NotImplementedError()

    async def render_async(self, **kwargs):
        """Render the template to an object.

        Keyword arguments are passed to the template."""
        raise NotImplementedError()

    def write(self, out: Union[Path, str, TextIO]):
        """Write the template source out.

        Args:
            out: A file name, or a file-like object.
        """
        if isinstance(out, (str, Path)):
            with open(out, "w") as f:
                f.write(self.source)
        else:
            out.write(self.source)


class JobStepConfigTemplate(FusedTemplate):
    """Template that generates a JobStepConfig.

    The extension for this type of template is `job_tmpl`.
    """

    def print(
        self,
        left: CoerceableToDatasetInput,
        right: CoerceableToDatasetInput,
        output: Optional[CoerceableToDatasetOutput] = None,
        **kwargs,
    ) -> None:
        print(
            self.render(
                left=left,
                right=right,
                output=output,
                **kwargs,
            ).print_code()
        )

    def code(
        self,
        left: CoerceableToDatasetInput,
        right: CoerceableToDatasetInput,
        output: Optional[CoerceableToDatasetOutput] = None,
        **kwargs,
    ) -> None:
        import IPython

        IPython.get_ipython().set_next_input(
            self.render(
                left=left,
                right=right,
                output=output,
                **kwargs,
            ).print_code(),
            replace=True,
        )

    def render(
        self,
        left: CoerceableToDatasetInput,
        right: CoerceableToDatasetInput,
        output: Optional[CoerceableToDatasetOutput] = None,
        **kwargs,
    ) -> AnyJobStepConfig:
        resolved_kwargs = {
            **self.defaults,
            "left": _object_to_dataset_input(left),
            "right": _object_to_dataset_input(right),
            "output": _object_to_dataset_output(output),
            **kwargs,
        }
        json = self.template.render(**resolved_kwargs)
        config = RootAnyJobStepConfig.model_validate_json(json).root
        return config

    async def render_async(
        self,
        left: CoerceableToDatasetInput,
        right: CoerceableToDatasetInput,
        output: Optional[CoerceableToDatasetOutput] = None,
        **kwargs,
    ) -> JobStepConfig:
        resolved_kwargs = {
            **self.defaults,
            "left": _object_to_dataset_input(left),
            "right": _object_to_dataset_input(right),
            "output": _object_to_dataset_output(output),
            **kwargs,
        }
        json = await self.template.render_async(**resolved_kwargs)
        config = RootAnyJobStepConfig.model_validate_json(json).root
        return config


class JobConfigTemplate(FusedTemplate):
    """Template that generates a JobConfig.

    The extension for this type of template is `job_tmpl`.
    """

    # TODO: print
    # TODO: code

    def render(
        self,
        input: Optional[CoerceableToDatasetInput] = None,
        left: Optional[CoerceableToDatasetInput] = None,
        right: Optional[CoerceableToDatasetInput] = None,
        output: Optional[CoerceableToDatasetOutput] = None,
        **kwargs,
    ) -> JobConfig:
        resolved_kwargs = {
            **self.defaults,
            "input": _object_to_dataset_input(input) if input else None,
            "left": _object_to_dataset_input(left),
            "right": _object_to_dataset_input(right),
            "output": _object_to_dataset_output(output),
            **kwargs,
        }
        json = self.template.render(**resolved_kwargs)
        config = JobConfig.model_validate_json(json)
        return config

    async def render_async(
        self,
        input: Optional[CoerceableToDatasetInput] = None,
        left: Optional[CoerceableToDatasetInput] = None,
        right: Optional[CoerceableToDatasetInput] = None,
        output: Optional[CoerceableToDatasetOutput] = None,
        **kwargs,
    ) -> JobConfig:
        resolved_kwargs = {
            **self.defaults,
            "input": _object_to_dataset_input(input) if input else None,
            "left": _object_to_dataset_input(left),
            "right": _object_to_dataset_input(right),
            "output": _object_to_dataset_output(output),
            **kwargs,
        }
        json = await self.template.render_async(**resolved_kwargs)
        config = JobConfig.model_validate_json(json)
        return config


class UdfTemplate(FusedTemplate):
    """Template that generates a UDF.

    The extension for this type of template is `udf_tmpl`.
    """

    def print(self, **kwargs) -> None:
        print(self.render(**kwargs).code)

    def code(self, **kwargs) -> None:
        import IPython

        IPython.get_ipython().set_next_input(self.render(**kwargs).code, replace=True)

    def render(self, **kwargs) -> AnyBaseUdf:
        resolved_kwargs = {
            **self.defaults,
            **kwargs,
        }
        json = self.template.render(**resolved_kwargs)
        return RootAnyBaseUdf.model_validate_json(json).root

    async def render_async(self, **kwargs) -> AnyBaseUdf:
        resolved_kwargs = {
            **self.defaults,
            **kwargs,
        }
        json = await self.template.render_async(**resolved_kwargs)
        return RootAnyBaseUdf.model_validate_json(json).root


class Templates:
    """A collection of templates.

    Each template can be accessed with subscript operator or dot operator, e.g.
    `templates["template_name"]` or `templates.template_name`. Templates may be
    registered on this object using the subscript operator.

    Example:

        ```py
        fused.templates.udfs['my_udf_template'] = my_udf
        ```

    Registered templates always take priority over templates loaded from files,
    and over built-in templates.
    """

    _env: Environment
    _choice_loader: ChoiceLoader
    _name: str
    _template_class: Type[FusedTemplate]
    _defaults: Dict[str, Any]
    _extension: str
    _registered: Dict[str, str]

    def __init__(
        self,
        name: str,
        template_class: Type,
        extension: str,
        defaults: Optional[Dict[str, Any]] = None,
    ):
        self._name = name
        self._template_class = template_class
        self._defaults = defaults if defaults is not None else {}
        self._extension = extension
        self._registered = {}
        self._choice_loader = ChoiceLoader([])
        self._reset_loaders()
        self._env = _create_env(self._choice_loader)
        self._env.filters["embedpy"] = _embedpy_filter

    def _repr_html_(self):
        return fused_templates_repr(self._name, self._all_templates_by_loader())

    def __getitem__(self, key: str) -> FusedTemplate:
        template_key = self._name_to_template_key(key)
        source, _filename, _uptodate = self._env.loader.get_source(
            self._env, template_key
        )
        template = self._env.get_template(template_key)
        return self._template_class(
            template=template,
            name=key,
            source=source,
            defaults=self._defaults,
        )

    def __setitem__(self, key: str, value: Union[BaseModel, str]):
        if isinstance(value, BaseModel):
            value = value.json(indent=4)
        if not isinstance(value, str):
            raise TypeError(f"Template source has unexpected type {type(value)}")
        template_key = self._name_to_template_key(key)
        self._registered[template_key] = value

    def __delitem__(self, key: str):
        template_key = self._name_to_template_key(key)
        del self._registered[template_key]

    def __getattribute__(self, key):
        try:
            return super().__getattribute__(key)
        except AttributeError:
            try:
                return self[key]
            # Note that we need to raise an AttributeError, **not a TemplateNotFound** so that
            # IPython's _repr_html_ works here
            except TemplateNotFound:
                raise AttributeError(
                    f"object of type {type(self).__name__} has no attribute {key}"
                ) from None

    def list_templates(self) -> List[str]:
        """Returns a list of all template names in this collection."""
        return [
            name.rsplit(f".{self._extension}", maxsplit=1)[0]
            for name in self._env.list_templates(extensions=[self._extension])
        ]

    def load_files(self, where: Union[Path, str], **kwargs) -> "Templates":
        """Add a directory to the search path for templates.

        Loaded templates may override built-in templates, but will not take priority
        over templates registered on this object.

        Arguments:
            where: Path to a directory to load. The directory should have files with the extension for this registry, e.g. `job_tmpl` for `fused.templates.jobs`.
        """
        # Index 1 puts the loaded templates *after* the DictLoader for the registered templates
        # TODO: Maybe add an option for loading a Python source file directly as a header
        self._choice_loader.loaders.insert(1, FileSystemLoader(where, **kwargs))
        return self

    def load_url(
        self,
        url: str,
        name: str,
        **request_kwargs,
    ) -> "Templates":
        """Add a URL to the search path for templates.

        Loaded templates may override built-in templates, but will not take priority
        over templates registered on this object.

        Arguments:
            url: URL to load from. A single template should be returned by the URL.
            name: Name to save the loaded template as.
        """
        # TODO: Support loading directly from gist here
        r = requests.get(
            url=url,
            timeout=OPTIONS.request_timeout,
            **request_kwargs,
        )
        raise_for_status(r)

        key = self._name_to_template_key(name)
        source = r.text

        self._choice_loader.loaders.insert(1, UrlLoader({key: source}, url=url))
        return self

    def from_object(
        self, obj: BaseModel, name: str, register: bool = False
    ) -> FusedTemplate:
        """Create a template from an object.

        Args:
            obj: The object to create a template from. It should be of the appropriate type for this registry, e.g. a `AnyBaseUdf` for the UDFs registry.
            name: The name of the template to create.
            register: If True, store the resulting template in this registry. Defaults to False.
        """
        if register:
            self[name] = obj
            return self[name]

        # Otherwise, we need to specifically create the template
        source = obj.json(indent=4)
        template = _create_env(DictLoader({name: source})).get_template(name)
        return self._template_class(
            template=template,
            name=name,
            source=source,
            defaults=self._defaults,
        )

    def _all_templates(self) -> Dict[str, FusedTemplate]:
        return {name: self[name] for name in self.list_templates()}

    def _all_templates_by_loader(self) -> Dict[str, Tuple[FusedTemplate, BaseLoader]]:
        # This supports finding all the templates and returning them with their loader
        # for use in the repr
        results: Dict[str, Tuple[FusedTemplate, BaseLoader]] = {}
        for loader in self._choice_loader.loaders:
            all_possible_names = loader.list_templates()
            # We have to apply the extension filter our selves here
            names_with_extensions = [
                name.rsplit(".", maxsplit=1)
                for name in all_possible_names
                if "." in name
            ]
            names = [
                name_and_extension[0]
                for name_and_extension in names_with_extensions
                if name_and_extension[1] == self._extension
            ]
            for name in names:
                if name not in results:
                    # Note that we don't necessarily know that the template is loaded from
                    # this loader. Instead we are replicating the "choose first" behavior from
                    # Jinja.
                    results[name] = (self[name], loader)
        return results

    def _name_to_template_key(self, key: str) -> str:
        return f"{key}.{self._extension}"

    def _reset_loaders(self):
        self._choice_loader.loaders = [
            RegisteredLoader(self._registered),
            PackageLoader("fused_batch._static"),
        ]

    def __dir__(self) -> List[str]:
        # Provide method name lookup and completion. Only provide 'public' methods.
        # This enables autocompletion
        extra_attrs = set(self.list_templates())
        normal_dir = {name for name in dir(type(self)) if not name.startswith("_")}

        return sorted(normal_dir | extra_attrs)


def _create_env(loader: BaseLoader) -> Environment:
    return Environment(
        loader=loader,
        autoescape=select_autoescape(),
        undefined=StrictUndefined,
    )


def _embedpy_filter(code: str) -> str:
    assert code is not None, "embedpy filter called on None"
    assert isinstance(
        code, str
    ), f"embedpy filter called on a non-string (type was actually {type(code).__name__})"
    return json.dumps(code)[1:-1]


jobs = Templates(
    name="jobs", template_class=JobStepConfigTemplate, extension="job_tmpl"
)
"""Templates for jobs (joins, maps, etc)"""
job = Templates(name="job", template_class=JobConfigTemplate, extension="job_tmpl")
"""Templates for jobs with one or more steps"""
udfs = Templates(name="udfs", template_class=UdfTemplate, extension="udf_tmpl")
"""Templates for UDFs"""



# ruff: noqa: F401

from fused._auth import AUTHORIZATION

from ._public_api import (
    delete,
    download,
    get,
    get_apps,
    get_udfs,
    job_cancel,
    job_get_exec_time,
    job_get_logs,
    job_get_status,
    job_print_logs,
    job_tail_logs,
    job_wait_for_job,
    list,
    sign_url,
    sign_url_prefix,
    upload,
    whoami,
)
from .api import FusedAPI
from .credentials import NotebookCredentials, access_token, auth_scheme, logout
from .docker_api import FusedDockerAPI

__all__ = [
    "FusedAPI",
    "NotebookCredentials",
    "delete",
    "download",
    "get",
    "get_apps",
    "get_udfs",
    "job_cancel",
    "job_get_exec_time",
    "job_get_logs",
    "job_get_status",
    "job_print_logs",
    "job_tail_logs",
    "job_wait_for_job",
    "list",
    "sign_url",
    "sign_url_prefix",
    "upload",
    "whoami",
]



from __future__ import annotations

import shlex
from typing import TYPE_CHECKING, List, Optional

import click
from IPython.core.magic import Magics, line_magic, magics_class

from fused_batch._global_api import get_api
from fused_batch.models.schema import Field, Schema

if TYPE_CHECKING:
    from IPython.core.interactiveshell import InteractiveShell


@click.command()
@click.argument("udf_name", type=str, required=True)
def load_udf_line_parser():
    """Define a user-defined function template"""
    pass


@magics_class
class FusedMagics(Magics):
    fused_shell: InteractiveShell

    def __init__(self, shell: InteractiveShell):
        self.fused_shell = shell
        super().__init__(shell)

    @staticmethod
    def parse_fields(fields: List[str]) -> Optional[Schema]:
        if len(fields) == 0:
            return None

        parsed_fields: List[Field] = []
        for field in fields:
            split = field.split(":", maxsplit=1)
            if len(split) == 1:
                raise NotImplementedError()

            name = split[0]
            dtype = split[1]
            parsed_field = Field(name=name, type=dtype)  # type: ignore
            parsed_fields.append(parsed_field)

        return Schema(fields=parsed_fields)

    @line_magic
    def load_job(self, line: str) -> None:
        job_name = line.strip()
        job_step_config = get_api().get_job_config(job_name)._to_job_step_config()
        job_step_config.render()

    @line_magic
    def load_udf(self, line: str) -> None:
        ctx = click.Context(load_udf_line_parser)
        parser = load_udf_line_parser.make_parser(ctx)
        shell_split_args = self._split_args(line)
        values, _args, _order = parser.parse_args(shell_split_args)

        help_text = load_udf_line_parser.get_help(ctx)

        # Handle --help.
        should_show_help = "help" in values.keys() and values["help"] is True
        if should_show_help:
            print(help_text)
            return

        # We don't use .get() here because udf_name defaults to None
        udf_name = values.get("udf_name") or ""
        udf_name = udf_name.strip()
        if udf_name == "":
            raise ValueError(
                f"No name was passed for the user-defined function.\n\n{help_text}"
            )

        udf = eval(udf_name, self.fused_shell.user_global_ns, self.fused_shell.user_ns)
        udf.render()

    def _split_args(self, line: str) -> List[str]:
        """Split arguments using shell style splitting.

        This is needed beyond splitting on whitespace so that arguments
        containing whitespace can be passed in. It also makes comments (#)
        behave as expected."""
        return shlex.split(line, comments=True)


def load_ipython_extension(ipython: InteractiveShell):
    magics = FusedMagics(ipython)
    ipython.register_magics(magics)



from __future__ import annotations

import uuid
from functools import partial
from html import escape
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple

from jinja2 import BaseLoader, FileSystemLoader, PackageLoader
from xarray.core.options import _get_boolean_with_default

from fused_batch._formatter.common import icon
from fused_batch._templates.loaders import RegisteredLoader, UrlLoader

if TYPE_CHECKING:
    from fused_batch._templates import FusedTemplate


def collapsible_section(
    name,
    inline_details: str = "",
    details: str = "",
    n_items: Optional[int] = None,
    enabled: bool = True,
    collapsed: bool = False,
):
    # "unique" id to expand/collapse the section
    data_id = "section-" + str(uuid.uuid4())

    has_items = n_items is not None and n_items
    n_items_span = "" if n_items is None else f" <span>({n_items})</span>"
    enabled_str = "" if enabled and has_items else "disabled"
    collapsed_str = "" if collapsed or not has_items else "checked"
    tip = " title='Expand/collapse section'" if enabled else ""

    return (
        f"<input id='{data_id}' class='xr-section-summary-in' "
        f"type='checkbox' {enabled_str} {collapsed_str}>"
        f"<label for='{data_id}' class='xr-section-summary' {tip}>"
        f"{name}:{n_items_span}</label>"
        f"<div class='xr-section-inline-details'>{inline_details}</div>"
        f"<div class='xr-section-details'>{details}</div>"
    )


def _mapping_section(
    mapping,
    name: str,
    details_func: Callable[[Any], str],
    max_items_collapse,
    expand_option_name,
    enabled: bool = True,
):
    n_items = len(mapping)
    expanded = _get_boolean_with_default(
        expand_option_name, n_items < max_items_collapse
    )
    collapsed = not expanded

    return collapsible_section(
        name,
        details=details_func(mapping),
        n_items=n_items,
        enabled=enabled,
        collapsed=collapsed,
    )


def summarize_template(
    template_name: str,
    template: FusedTemplate,
    loader: Optional[BaseLoader] = None,
    *,
    is_index: bool = False,
    dtype=None,
):
    cssclass_idx = " class='xr-has-index'" if is_index else ""
    code_lines = template.source.splitlines()
    dims_str = f"LOCs:{len(code_lines)}"
    name = escape(str(template_name))
    dtype = ""

    # "unique" ids required to expand/collapse subsections
    attrs_id = "attrs-" + str(uuid.uuid4())
    data_id = "data-" + str(uuid.uuid4())
    chunk_repr_disabled = "disabled" if template.source is None else ""

    # Convert DataFrame to HTML udf
    # TODO: escape html characters # escape(inline_variable_array_repr(variable, 35))
    if isinstance(loader, FileSystemLoader):
        preview = f"files {loader.searchpath}"
    elif isinstance(loader, PackageLoader):
        preview = f"package {loader.package_name}"
    elif isinstance(loader, UrlLoader):
        preview = f"url {loader.url}"
    elif isinstance(loader, RegisteredLoader):
        preview = "registered"
    else:
        preview = "unknown source"
    defaults = [f"<li>{k}: <code>{v}</code></li>" for k, v in template.defaults.items()]

    # if isinstance(table.parent.inputs[0], fused.models.internal.dataset.DatasetInput):
    attrs_ul = f"""
    <h4 style="margin-left: 20px;">Defaults</h4>
    <ul style="margin-top: 0; margin-bottom: 0px;">
    {''.join(defaults)}
    </ul>
     """  # summarize_attrs(var.attrs)
    rendered_code = "\n".join(
        [f'<span class="fused-udf-body">{line}</span>' for line in code_lines]
    )
    data_repr = f"""<div class="fused-udf-body-wrapper"><pre class="fused-udf-body"><code>{rendered_code}</code></pre></div>"""
    # data_repr = f'<div align="center" style="overflow-x:auto;">123</div>'
    # else:
    #     data_repr = ""
    # data_repr = ""
    attrs_icon = icon("icon-file-text2")
    code_icon = icon("icon-code")

    return (
        f"<div class='xr-var-name'><span{cssclass_idx}>{name}</span></div>"
        f"<div class='xr-var-dims'>{dims_str}</div>"
        f"<div class='xr-var-dtype'>{dtype}</div>"
        f"<div class='xr-var-preview xr-preview'>{preview}</div>"
        f"<input id='{attrs_id}' class='xr-var-attrs-in' "
        f"type='checkbox'>"
        f"<label for='{attrs_id}' title='Show/Hide attributes'>"
        f"{attrs_icon}</label>"
        f"<input id='{data_id}' class='xr-var-data-in' type='checkbox' {chunk_repr_disabled}>"
        f"<label for='{data_id}' title='Show/Hide data repr'>"
        f"{code_icon}</label>"
        f"<div class='xr-var-attrs'>{attrs_ul}</div>"
        f"<div class='xr-var-data'>{data_repr}</div>"
    )


def summarize_templates(
    templates: Dict[str, Tuple[FusedTemplate, Optional[BaseLoader]]]
):
    li_items = []
    for template_name, template_and_loader in templates.items():
        template, loader = template_and_loader
        li_content = summarize_template(
            template_name=template_name,
            template=template,
            loader=loader,
            is_index=False,
        )
        li_items.append(f"<li class='xr-var-item'>{li_content}</li>")

    vars_li = "".join(li_items)

    return f"<ul class='xr-var-list'>{vars_li}</ul>"


template_section = partial(
    _mapping_section,
    name="Templates",
    details_func=summarize_templates,
    max_items_collapse=25,
    expand_option_name="display_expand_coords",
)



from __future__ import annotations

import warnings
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    TextIO,
    Tuple,
    Union,
)

import geopandas as gpd
from pydantic import BaseModel, StrictBool, StrictStr

import fused_batch._public_api
from fused_batch._formatter.formatter_project import fused_project_repr
from fused_batch._global_api import get_api
from fused_batch._options import options as OPTIONS
from fused_batch._str_utils import append_url_part, table_to_name
from fused_batch.models.api import (
    Dataset,
    GeospatialPartitionJobStepConfig,
    JoinJobStepConfig,
    MapJobStepConfig,
    NonGeospatialPartitionJobStepConfig,
    Table,
)
from fused_batch.models.api.job import GDALOpenConfig, JoinType
from fused_batch.models.internal.dataset import (
    DatasetInputV2,
    DatasetInputV2Table,
    DatasetInputV2Type,
    DatasetOutputV2,
)
from fused_batch.models.schema import Schema
from fused_batch.models.udf import AnyBaseUdf
from fused_batch.models.urls import DatasetUrl
from fused_batch.warnings import (
    FusedIgnoredWarning,
    FusedPathWarning,
    FusedRefreshingWarning,
)


class Project(BaseModel):
    """A project represents a collection of tables or directories of files in blob storage.

    Tables and folders under this project may be accessed with attribute or subscript operator.
    For example, all of the following will have the same result:

    ```py
    project.tables['my_table_name']

    project.my_table_name

    project['my_table_name']
    ```
    """

    root_base_path: DatasetUrl
    """The base path of the overall project."""
    base_path: DatasetUrl
    """The base path of the project folder, which tables in it are relative to. This may be different
    than `root_base_path` if this Project instance is a folder (sub-project) of an overall project."""

    tables: Dict[StrictStr, Table]
    """Tables in this project."""
    folders: Dict[StrictStr, Project]
    """Project folders in this project."""
    virtual_folders: List[StrictStr]
    """Project folders in this project that have not been materialized.

    Accessing one of these through `project[virtual_folder_name]` will result in automatic
    loading of that folder to a `Project` instance.
    """

    # The following are settings that were used to open the project and the user shouldn't
    # need to change. Modifying them after opening a project will only take effect in case
    # of `refresh()`.
    fetch_table_metadata: Optional[StrictBool] = None
    fetch_minimal_table_metadata: Optional[StrictBool] = None
    fetch_samples: Optional[StrictBool] = None

    def _getitem_internal(self, key: str) -> Union[Project, Table, None]:
        if key in self.tables:
            # If a table and a folder are both specified, the table takes precedence
            return self.tables[key]
        elif key in self.folders:
            return self.folders[key]
        elif key in self.virtual_folders:
            return self._materialize_virtual_folder(key)
        else:
            # TODO: Allow specifying random paths
            return None

    def __getitem__(self, key: str) -> Union[Project, Table]:
        ret = self._getitem_internal(key)
        if ret is None:
            if OPTIONS.open.auto_refresh_project and not key.startswith("_"):
                # Do not do the refresh for keys that start with _ because IPython
                # will probe a number of attributes (such as the _ipython_canary...
                # attribute) when rendering an object. This will make the repr take a
                # very long time.
                warnings.warn(
                    f"Refreshing project because key {key} was not found",
                    FusedRefreshingWarning,
                )
                self.refresh()

                # Try again:
                ret = self._getitem_internal(key)

            if ret is None:
                # Check again because ret may have been modified by auto_refresh_project,
                # above.
                raise KeyError(f"Key {key} not found")
        return ret

    def __getattribute__(self, key) -> Union[Project, Table]:
        try:
            return super().__getattribute__(key)
        except AttributeError:
            try:
                return self[key]
            # Note that we need to raise an AttributeError, **not a KeyError** so that
            # IPython's _repr_html_ works here
            except KeyError:
                raise AttributeError(
                    f"object of type {type(self).__name__} has no attribute {key}"
                ) from None

    def _repr_html_(self):
        return fused_project_repr(self)

    def __dir__(self) -> List[str]:
        # Provide method name lookup and completion. Only provide 'public'
        # methods.
        # This enables autocompletion
        # Pydantic methods to remove in __dir__
        PYDANTIC_METHODS = {
            "Config",
            "construct",
            "copy",
            "from_orm",
            "json",
            "parse_file",
            "parse_obj",
            "schema",
            "schema_json",
            "update_forward_refs",
            "validate",
        }
        EXTRA_METHODS = {
            "fetch_table_metadata",
            "fetch_minimal_table_metadata",
            "fetch_samples",
        }

        extra_attrs = set(self.tables.keys()) | set(
            self.folders.keys() | set(self.virtual_folders)
        )
        normal_dir = {
            name
            for name in dir(type(self))
            if (
                not name.startswith("_")
                and name not in PYDANTIC_METHODS
                and name not in EXTRA_METHODS
            )
        }
        pydantic_fields = set(self.model_fields.keys())
        return sorted(normal_dir | pydantic_fields | extra_attrs)

    def _ipython_key_completions_(self) -> List[str]:
        # https://ipython.readthedocs.io/en/stable/config/integrating.html#tab-completion
        return sorted(
            set(self.tables.keys())
            | set(self.folders.keys() | set(self.virtual_folders))
        )

    def _materialize_all_virtual_folders(self) -> Iterable[Project]:
        to_materialize = [*self.virtual_folders]
        for key in to_materialize:
            yield self._materialize_virtual_folder(key)

    def _materialize_virtual_folder(self, folder_name: str) -> Project:
        self.virtual_folders.remove(folder_name)
        new_folder = open_project(
            append_url_part(self.base_path, folder_name),
            _fetch_minimal_table_metadata=self.fetch_minimal_table_metadata,
            fetch_table_metadata=self.fetch_table_metadata,
            fetch_samples=self.fetch_samples,
        )
        self.folders[folder_name] = new_folder
        return new_folder

    def tree(self, file: TextIO = None) -> None:
        """Print a tree representation of this project.

        Args:
            file: File-like object to write to. Defaults to `None` for `sys.stdout`.
        """

        def _print_tree(project: Project, indent: int):
            prefix = ("|   " * (indent - 1)) + "|-- "
            for name in project.tables.keys():
                print(f"{prefix}{name}", file=file)
            for name, folder in project.folders.items():
                print(f"{prefix}{name}", file=file)
                _print_tree(folder, indent + 1)
            for name, folder in project.virtual_folders:
                print(f"{prefix}{name} (virtual)", file=file)

        print(self.base_path, file=file)
        _print_tree(self, 1)

    def refresh(
        self,
        *,
        fetch_table_metadata: Optional[bool] = None,
        fetch_samples: Optional[bool] = None,
        _fetch_minimal_table_metadata: Optional[bool] = None,
    ) -> Project:
        """Returns this project with updated metadata

        Keyword args, if specified, will change how the project loads metadata. This can be used to reload a project
        with metadata, after it is initially loaded without metadata.

        Keyword Args:
            fetch_table_metadata: If True, fetch metadata on each table.
            fetch_samples: If True, fetch sample on each table.
        """
        if fetch_table_metadata is not None:
            self.fetch_table_metadata = fetch_table_metadata
        if fetch_samples is not None:
            self.fetch_samples = fetch_samples
        if _fetch_minimal_table_metadata is not None:
            self.fetch_minimal_table_metadata = _fetch_minimal_table_metadata

        new_self = open_project(
            self.base_path,
            _fetch_minimal_table_metadata=self.fetch_minimal_table_metadata,
            fetch_table_metadata=self.fetch_table_metadata,
            fetch_samples=self.fetch_samples,
        )
        self.tables = new_self.tables
        self.folders = new_self.folders
        self.virtual_folders = new_self.virtual_folders
        return self

    def path(self, path: str) -> str:
        """Returns the path to an item under this project."""
        return append_url_part(self.base_path, path)

    def project(self, path: str) -> Project:
        """Open a subproject of this project."""
        if path in self.folders:
            return self.folders[path]
        elif path in self.virtual_folders:
            return self._materialize_virtual_folder(path)
        else:
            return Project(
                base_path=self.path(path),
                tables={},
                folders={},
                fetch_minimal_table_metadata=self.fetch_minimal_table_metadata,
                fetch_table_metadata=self.fetch_table_metadata,
                fetch_samples=self.fetch_samples,
            )

    def open_table(
        self,
        path: Union[str, DatasetOutputV2],
        *,
        fetch_samples: Optional[bool] = None,
    ) -> Table:
        """Open a Table object given a path to the root of the table

        Args:
            path: The path to the root of the table on remote storage

        Keyword Args:
            fetch_samples: If True, fetch sample on each table when getting dataset metadata.

        Example:

            table = project.open_table("path/to/dataset/table/")

        Returns:
            A Table object
        """
        if isinstance(path, str):
            path = self.path(path)
        elif isinstance(path, DatasetOutputV2):
            path = path.url

        return fused_batch._public_api.open_table(
            path=path,
            fetch_samples=fetch_samples,
        )

    def ingest(
        self,
        input: Union[str, Sequence[str], Path, gpd.GeoDataFrame],
        output: Optional[str] = None,
        *,
        output_metadata: Optional[str] = None,
        schema: Optional[Schema] = None,
        file_suffix: Optional[str] = None,
        load_columns: Optional[Sequence[str]] = None,
        remove_cols: Optional[Sequence[str]] = None,
        explode_geometries: bool = False,
        drop_out_of_bounds: Optional[bool] = None,
        partitioning_method: Literal["area", "length", "coords", "rows"] = "rows",
        partitioning_maximum_per_file: Union[int, float, None] = None,
        partitioning_maximum_per_chunk: Union[int, float, None] = None,
        partitioning_max_width_ratio: Union[int, float] = 2,
        partitioning_max_height_ratio: Union[int, float] = 2,
        partitioning_force_utm: Literal["file", "chunk", None] = "chunk",
        partitioning_split_method: Literal["mean", "median"] = "mean",
        subdivide_method: Literal["area", None] = None,
        subdivide_start: Optional[float] = None,
        subdivide_stop: Optional[float] = None,
        split_identical_centroids: bool = True,
        target_num_chunks: int = 5000,
        lonlat_cols: Optional[Tuple[str, str]] = None,
        gdal_config: Union[GDALOpenConfig, Dict[str, Any], None] = None,
    ) -> GeospatialPartitionJobStepConfig:
        """Ingest a dataset into the Fused partitioned format.

        Args:
            input: A GeoPandas `GeoDataFrame` or a path to file or files on S3 to ingest. Files may be Parquet or another geo data format.
            output: Location on S3 to write the `main` table to.
            output_metadata: Location on S3 to write the `fused` table to.
            schema: Schema of the data to be ingested. This is optional and will be inferred from the data if not provided.
            file_suffix: filter which files are used for ingestion. If `input` is a directory on S3, all files under that directory will be listed and used for ingestion. If `file_suffix` is not None, it will be used to filter paths by checking the trailing characters of each filename. E.g. pass `file_suffix=".geojson"` to include only GeoJSON files inside the directory.
            load_columns: Read only this set of columns when ingesting geospatial datasets. Defaults to all columns.
            remove_cols: The named columns to drop when ingesting geospatial datasets. Defaults to not drop any columns.
            explode_geometries: Whether to unpack multipart geometries to single geometries when ingesting geospatial datasets, saving each part as its own row. Defaults to `False`.
            drop_out_of_bounds: Whether to drop geometries outside of the expected WGS84 bounds. Defaults to True.
            partitioning_method: The method to use for grouping rows into partitions. Defaults to `"rows"`.
                - `"area"`: Construct partitions where all contain a maximum total area among geometries.
                - `"length"`: Construct partitions where all contain a maximum total length among geometries.
                - `"coords"`: Construct partitions where all contain a maximum total number of coordinates among geometries.
                - `"rows"`: Construct partitions where all contain a maximum number of rows.
            partitioning_maximum_per_file: Maximum value for `partitioning_method` to use per file. If `None`, defaults to _1/10th_ of the total value of `partitioning_method`. So if the value is `None` and `partitioning_method` is `"area"`, then each file will be have no more than 1/10th the total area of all geometries. Defaults to `None`.
            partitioning_maximum_per_chunk: Maximum value for `partitioning_method` to use per chunk. If `None`, defaults to _1/100th_ of the total value of `partitioning_method`. So if the value is `None` and `partitioning_method` is `"area"`, then each file will be have no more than 1/100th the total area of all geometries. Defaults to `None`.
            partitioning_max_width_ratio: The maximum ratio of width to height of each partition to use in the ingestion process. So for example, if the value is `2`, then if the width divided by the height is greater than `2`, the box will be split in half along the horizontal axis. Defaults to `2`.
            partitioning_max_height_ratio: The maximum ratio of height to width of each partition to use in the ingestion process. So for example, if the value is `2`, then if the height divided by the width is greater than `2`, the box will be split in half along the vertical axis. Defaults to `2`.
            partitioning_force_utm: Whether to force partitioning within UTM zones. If set to `"file"`, this will ensure that the centroid of all geometries per _file_ are contained in the same UTM zone. If set to `"chunk"`, this will ensure that the centroid of all geometries per _chunk_ are contained in the same UTM zone. If set to `None`, then no UTM-based partitioning will be done. Defaults to "chunk".
            partitioning_split_method: How to split one partition into children. Defaults to `"mean"` (this may change in the future).
                - `"mean"`: Split each axis according to the mean of the centroid values.
                - `"median"`: Split each axis according to the median of the centroid values.
            subdivide_method: The method to use for subdividing large geometries into multiple rows. Currently the only option is `"area"`, where geometries will be subdivided based on their area (in WGS84 degrees).
            subdivide_start: The value above which geometries will be subdivided into smaller parts, according to `subdivide_method`.
            subdivide_stop: The value below which geometries will not be subdivided into smaller parts, according to `subdivide_method`. Recommended to be equal to subdivide_start. If `None`, geometries will be subdivided up to a recursion depth of 100 or until the subdivided geometry is rectangular.
            split_identical_centroids: If `True`, should split a partition that has
                identical centroids (such as if all geometries in the partition are the
                same) if there are more such rows than defined in "partitioning_maximum_per_file" and
                "partitioning_maximum_per_chunk".
            target_num_chunks: The target for the number of chunks if `partitioning_maximum_per_file` is None. Note that this number is only a _target_ and the actual number of files and chunks generated can be higher or lower than this number, depending on the spatial distribution of the data itself.
            lonlat_cols: Names of longitude, latitude columns to construct point geometries from.

                If your point columns are named `"x"` and `"y"`, then pass:

                ```py
                fused.ingest(
                    ...,
                    lonlat_cols=("x", "y")
                )
                ```

                This only applies to reading from Parquet files. For reading from CSV files, pass options to `gdal_config`.

            gdal_config: Configuration options to pass to GDAL for how to read these files. For all files other than Parquet files, Fused uses GDAL as a step in the ingestion process. For some inputs, like CSV files or zipped shapefiles, you may need to pass some parameters to GDAL to tell it how to open your files.

                This config is expected to be a dictionary with up to two keys:

                - `layer`: `str`. Define the layer of the input file you wish to read when the source contains multiple layers, as in GeoPackage.
                - `open_options`: `Dict[str, str]`. Pass in key-value pairs with GDAL open options. These are defined on each driver's page in the GDAL documentation. For example, the [CSV driver](https://gdal.org/drivers/vector/csv.html) defines [these open options](https://gdal.org/drivers/vector/csv.html#open-options) you can pass in.

                For example, if you're ingesting a CSV file with two columns
                `"longitude"` and `"latitude"` denoting the coordinate information, pass

                ```py
                fused.ingest(
                    ...,
                    gdal_config={
                        "open_options": {
                            "X_POSSIBLE_NAMES": "longitude",
                            "Y_POSSIBLE_NAMES": "latitude",
                        }
                    }
                )
                ```

        Returns:
            Configuration object describing the ingestion process. Call `.execute` on this object to start a job.


        Examples:
            ```py
            job = fused.ingest(
                input="https://www2.census.gov/geo/tiger/TIGER_RD18/STATE/06_CALIFORNIA/06/tl_rd22_06_bg.zip",
                output="s3://fused-sample/census/ca_bg_2022/main/",
                output_metadata="s3://fused-sample/census/ca_bg_2022/fused/",
                explode_geometries=True,
                partitioning_maximum_per_file=2000,
                partitioning_maximum_per_chunk=200,
            ).execute()
            ```

        """
        # TODO: Support input being resolved
        project_output = self.path(output) if output else None
        project_output_metadata = (
            self.path(output_metadata) if output_metadata else None
        )
        return fused_batch._public_api.ingest(
            input=input,
            output=project_output,
            output_metadata=project_output_metadata,
            schema=schema,
            file_suffix=file_suffix,
            load_columns=load_columns,
            remove_cols=remove_cols,
            explode_geometries=explode_geometries,
            drop_out_of_bounds=drop_out_of_bounds,
            lonlat_cols=lonlat_cols,
            partitioning_maximum_per_file=partitioning_maximum_per_file,
            partitioning_maximum_per_chunk=partitioning_maximum_per_chunk,
            partitioning_max_width_ratio=partitioning_max_width_ratio,
            partitioning_max_height_ratio=partitioning_max_height_ratio,
            partitioning_method=partitioning_method,
            partitioning_force_utm=partitioning_force_utm,
            partitioning_split_method=partitioning_split_method,
            subdivide_start=subdivide_start,
            subdivide_stop=subdivide_stop,
            subdivide_method=subdivide_method,
            split_identical_centroids=split_identical_centroids,
            target_num_chunks=target_num_chunks,
            gdal_config=gdal_config,
        )

    def ingest_nongeospatial(
        self,
        input: Union[str, Sequence[str], Path, gpd.GeoDataFrame],
        output: Optional[str] = None,
        *,
        output_metadata: Optional[str] = None,
        partition_col: Optional[str] = None,
        partitioning_maximum_per_file: int = 2_500_000,
        partitioning_maximum_per_chunk: int = 65_000,
    ) -> NonGeospatialPartitionJobStepConfig:
        """Ingest a dataset into the Fused partitioned format.

        Args:
            input: A GeoPandas `GeoDataFrame` or a path to file or files on S3 to ingest. Files may be Parquet or another geo data format.
            output: Location on S3 to write the `main` table to.
            output_metadata: Location on S3 to write the `fused` table to.
            partition_col: Partition along this column for nongeospatial datasets.
            partitioning_maximum_per_file: Maximum number of items to store in a single file. Defaults to 2,500,000.
            partitioning_maximum_per_chunk: Maximum number of items to store in a single file. Defaults to 65,000.

        Returns:

            Configuration object describing the ingestion process. Call `.execute` on this object to start a job.

        Examples:
            ```py
            job = fused.ingest_nongeospatial(
                input=gdf,
                output="s3://sample-bucket/file.parquet",
            ).execute()
            ```
        """
        project_output = self.path(output) if output else None
        project_output_metadata = (
            self.path(output_metadata) if output_metadata else None
        )
        return fused_batch._public_api.ingest_nongeospatial(
            input=input,
            output=project_output,
            output_metadata=project_output_metadata,
            partition_col=partition_col,
            partitioning_maximum_per_file=partitioning_maximum_per_file,
            partitioning_maximum_per_chunk=partitioning_maximum_per_chunk,
        )

    def map(
        self,
        dataset: Union[str, Dataset, Table],
        output_table: Optional[str] = None,
        udf: Union[AnyBaseUdf, None] = None,
        *,
        cache_locally: bool = False,
    ) -> MapJobStepConfig:
        """Construct a `map` config from this Dataset

        Args:
            output_table: Where to save the output of this operation. Defaults to `None`, which will not save the output.
            udf: A user-defined function to run in this map. Defaults to None.

        Keyword Args:
            tables: The attribute tables to include in the map reduce. Defaults to ("main",).
            cache_locally: Advanced: whether to cache all the partitions locally in the map job. Defaults to False.

        Returns:
            An object describing the map configuration.
        """
        if isinstance(dataset, str):
            dataset = self.open_table(
                dataset,
                fetch_samples=False,
            )

        return dataset.map(
            output_table=output_table,
            udf=udf,
            cache_locally=cache_locally,
        )

    def join(
        self,
        dataset: Union[str, Dataset, Table],
        other: Union[str, Dataset, Table],
        output_table: Optional[str] = None,
        udf: Union[AnyBaseUdf, None] = None,
        *,
        how: Union[JoinType, Literal["left", "inner"]] = "inner",
        left_cache_locally: bool = False,
        right_cache_locally: bool = False,
        buffer_distance: Optional[float] = None,
    ) -> JoinJobStepConfig:
        """Construct a join config from two datasets

        Args:
            other: The other Dataset object to join on
            output_table: Where to save the output of this operation. Defaults to `None`, which will not save the output.
            udf: The user-defined function to run in the join

        Keyword Args:
            how: The manner of doing the join. Currently  Defaults to "inner".
            left_tables: The names of the attribute tables on the left side to include in the join.
            right_tables: The names of the attribute tables on the left side to include in the join.
            left_cache_locally: Whether to cache the left dataset locally in the join. Defaults to False.
            right_cache_locally: Whether to cache the right dataset locally in the join. Defaults to False.
            buffer_distance: The size of the buffer (in meters) on the left table to use during the join. Defaults to None.

        Examples:
            ```py
            import fused

            left_dataset = fused.open("s3://bucket/path/to/dataset")
            other_dataset = fused.open("s3://bucket/path/to/dataset")
            join_config = left_dataset.join(other_dataset)
            ```

        Returns:
            An object describing the join configuration.
        """
        if isinstance(dataset, str):
            dataset = self.open_table(dataset, fetch_samples=False)

        if isinstance(other, str):
            other = self.open_table(other, fetch_samples=False)

        return dataset.join(
            other=other,
            output_table=output_table,
            udf=udf,
            how=how,
            left_cache_locally=left_cache_locally,
            right_cache_locally=right_cache_locally,
            buffer_distance=buffer_distance,
        )

    def delete(
        self,
        path: str,
        max_deletion_depth: Union[int, Literal["unlimited"]] = 2,
    ) -> bool:
        """Delete the files at the path.

        Args:
            path: Directory or file to delete, like `my-old-table/`
            max_deletion_depth: If set (defaults to 2), the maximum depth the operation will recurse to.
                                This option is to help avoid accidentally deleting more data that intended.
                                Pass `"unlimited"` for unlimited.
        """
        project_path = self.path(path) if path is not None else self.base_path
        return fused_batch._public_api.delete(
            project_path, max_deletion_depth=max_deletion_depth
        )

    def list(self, path: Optional[str] = None) -> List[str]:
        """List the files at the path.

        Args:
            path: Parent directory, like `table_name`. Defaults to None to list the root of the project.

        Returns:
            A list of paths as URLs
        """
        project_path = self.path(path) if path is not None else self.base_path
        return fused_batch._public_api.list(project_path)

    def get(self, path: str) -> bytes:
        """Download the contents at the path to memory.

        Args:
            path: Path to a file, like `table_name/file.parquet`

        Returns:
            bytes of the file
        """
        project_path = self.path(path)
        return fused_batch._public_api.get(project_path)

    def download(self, path: str, local_path: Union[str, Path]) -> None:
        """Download the contents at the path to disk.

        Args:
            path: Path to a file, like `table_name/file.parquet`
            local_path: Path to a local file.
        """
        project_path = self.path(path)
        return fused_batch._public_api.download(project_path, local_path=local_path)

    def sign_url(self, path: str) -> str:
        """Create a signed URL to access the path.

        This function may not check that the file represented by the path exists.

        Args:
            path: Path to a file, like `table_name/file.parquet`

        Returns:
            HTTPS URL to access the file using signed access.
        """
        project_path = self.path(path)
        return fused_batch._public_api.sign_url(project_path)

    def sign_url_prefix(self, path: str) -> Dict[str, str]:
        """Create signed URLs to access all blobs under the path.

        Args:
            path: Path to a prefix, like `table_name/`

        Returns:
            Dictionary mapping from blob store key to signed HTTPS URL.
        """
        project_path = self.path(path)
        return fused_batch._public_api.sign_url_prefix(project_path)

    def _table_to_v2_table(
        self,
        table: Union[Table, str],
        *,
        read_sidecar: Union[Sequence[str], bool] = False,
    ) -> DatasetInputV2Table:
        table_url = table.url if isinstance(table, Table) else self.path(table)
        table_name = table_to_name(table_url)
        read_sidecar_bool = read_sidecar is not None and (
            read_sidecar
            if isinstance(read_sidecar, bool)
            else (table_name in read_sidecar)
        )
        return DatasetInputV2Table(
            url=table_url,
            read_sidecar_files=read_sidecar_bool,
        )

    def _create_table_objs(
        self,
        tables: Iterable[Union[Table, str]],
        read_sidecar: Union[Sequence[str], bool] = False,
    ):
        ret = [self._table_to_v2_table(t, read_sidecar=read_sidecar) for t in tables]

        if not isinstance(read_sidecar, bool):
            unapplied_sidecar_table_names: List[str] = []
            applied_sidecar_table_names = set(
                [table_to_name(t.url) for t in ret if t.read_sidecar_files is True]
            )
            for sidecar_table_name in read_sidecar:
                if sidecar_table_name not in applied_sidecar_table_names:
                    unapplied_sidecar_table_names.append(sidecar_table_name)

            if unapplied_sidecar_table_names:
                warnings.warn(
                    f"Some table names to read sidecars from were not applied: {unapplied_sidecar_table_names}",
                    FusedIgnoredWarning,
                )

        return ret

    def sel(
        self,
        tables: Union[Iterable[Union[Table, str]], Table, str, None] = None,
        *,
        read_sidecar: Union[Sequence[str], bool] = False,
        how: Optional[Union[str, DatasetInputV2Type]] = None,
        **kwargs,
    ) -> DatasetInputV2:
        """Create a job input that zips or unions tables together

        Args:
            tables: The names of tables to include in the input, e.g. `["table_0", "table_1", "table_5"]`.

        Keyword Args:
            read_sidecar: Whether to read sidecar information, either a sequence of table names (i.e. the last part
                          of the table path) to read it from or a boolean which will be applied to all tables (default False).
            how: The operation used to combine multiple input tables. This may be either `"zip"` or `"union"`.
                 By default this will be `"zip"` when `tables` is specified, `"union"` otherwise. This corresponds
                 with `fused.zip_tables` and `fused.union_tables` respectively.
        """
        default_how = DatasetInputV2Type.ZIP

        if len(kwargs) > 1 or (len(kwargs) == 1 and tables is not None):
            raise ValueError(
                "Too many keyword arguments passed in to fused.sel. There should only be a single `partition_name=values` argument."
            )

        if len(kwargs) == 0 and tables is None:
            raise ValueError(
                "No tables selected. Pass `tables=values` or `partition_name=values`."
            )

        if len(kwargs) == 1:
            default_how = DatasetInputV2Type.UNION

            assert tables is None
            kv_name = next(iter(kwargs.keys()))
            if isinstance(kwargs[kv_name], str):
                tables = [f"{kv_name}={kwargs[kv_name]}"]
            else:
                tables = [f"{kv_name}={kv_value}" for kv_value in kwargs[kv_name]]
        elif isinstance(tables, str) or isinstance(tables, Table):
            tables = [tables]

        if how is None:
            how = default_how

        if isinstance(how, str):
            how = DatasetInputV2Type(how)

        input = DatasetInputV2(
            tables=self._create_table_objs(tables=tables, read_sidecar=read_sidecar),
            operation=how,
        )
        input._project_url = self.root_base_path
        return input

    def isel(
        self,
        tables: Union[Iterable[int], int],
        *,
        read_sidecar: Union[Sequence[bool], bool] = False,
        how: Optional[Union[str, DatasetInputV2Type]] = None,
    ) -> DatasetInputV2:
        """Create a job input that zips or unions tables together, by their integer index. Tables
        are implicitly ordered by name.

        Args:
            tables: The index of tables to include in the input, e.g. `[0, 1, 5]`.

        Keyword Args:
            read_sidecar: Whether to read sidecar information, either a sequence of table names (i.e. the last part
                          of the table path) to read it from or a boolean which will be applied to all tables (default False).
            how: The operation used to combine multiple input tables. This may be either `"zip"` or `"union"`.
                 This corresponds with `fused.zip_tables` and `fused.union_tables` respectively. Defaults to `"zip"`.
        """
        default_how = DatasetInputV2Type.ZIP

        if how is None:
            how = default_how

        if isinstance(how, str):
            how = DatasetInputV2Type(how)

        table_names = sorted(self.tables.keys())
        read_sidecar = (
            ([read_sidecar] * len(table_names))
            if isinstance(read_sidecar, bool)
            else read_sidecar
        )

        input = DatasetInputV2(
            tables=self._create_table_objs(tables=tables, read_sidecar=read_sidecar),
            operation=how,
        )
        input._project_url = self.root_base_path
        return input


def open_project(
    path: str,
    *,
    lazy: bool = False,
    fetch_table_metadata: Optional[bool] = None,
    fetch_samples: Optional[bool] = None,
    _fetch_minimal_table_metadata: Optional[bool] = None,
    _max_depth: Optional[int] = 1,
    _eager: bool = False,
) -> Project:
    """Open a project folder.

    Args:
        path: Path to the project folder, e.g. `"s3://bucket-name/project-name/"`
        lazy: If True, no metadata about the project is loaded.
        fetch_table_metadata: This is passed on to the `Table` open calls.
        fetch_samples: This is passed on to the `Table` open calls.
        _fetch_minimal_table_metadata: If True and fetch_table_metadata is also True,
                                       a reduced set of Table metadata will be fetched.
        _max_depth: Maximum depth of folders to load.
        _eager: If True, recursive calls will be made to materialize all virtual
                folders that `max_depth` would otherwise cause.
    """
    if lazy:
        return Project(base_path=path, tables=[], folders=[])

    # TODO: Other values are not supported because the code for adding virtual folders
    # needs to be improved for it. Both where the virtual folders are added in the tree
    # and how to imlpement "eager" DFS mode.
    assert _max_depth is None or _max_depth == 1, "max_depth may only be None or 1"

    api = get_api()
    folder = api.open_folder(
        path=path,
        fetch_minimal_table_metadata=_fetch_minimal_table_metadata,
        fetch_table_metadata=fetch_table_metadata,
        fetch_samples=fetch_samples,
        max_depth=_max_depth,
    )

    all_tables = folder.tables

    # Create a prefix tree for the project
    project_tree: Dict[str, Any] = {}
    resolved_path = path if folder.base_path is None else folder.base_path

    for table in sorted(all_tables, key=lambda ds: ds.url):
        table_path = table.url
        if table_path.startswith(resolved_path):
            table_path = table_path[len(resolved_path) :]
            path_parts = table_path.strip("/").split("/")

            curr_tree_node = project_tree
            do_not_set = False
            for path_part in path_parts[:-1]:
                if path_part not in curr_tree_node:
                    curr_tree_node[path_part] = {}
                elif isinstance(curr_tree_node[path_part], Table):
                    # Note: if there is both a dataset and a folder,
                    # it will be shadowed by the dataset
                    warnings.warn(
                        f"Table {table.url} is nested within another table",
                        FusedPathWarning,
                    )
                    do_not_set = True
                    break
                curr_tree_node = curr_tree_node[path_part]

            if not do_not_set:
                curr_tree_node[path_parts[-1]] = table
        else:
            warnings.warn(f"Table path {table_path} is invalid", FusedPathWarning)

    # In "eager" mode, the frontend does DFS to materialize folders. This reduces the
    # pressure of each backend call (hopefully to within a reasonable timeout) while
    # allowing the frontend to get the whole data.
    def _recursive_eager_materialize(pr: Project):
        # note that non-virtual folders are not recursed into, so this also requires
        # max_depth=1.
        for folder in pr._materialize_all_virtual_folders():
            _recursive_eager_materialize(folder)

    # Convert the prefix tree into a tree of Project objects
    def _recursive_create_project(
        base_path: str, tree, additional_folders: Optional[Sequence[str]] = None
    ):
        tables: Dict[str, Table] = {}
        folders: Dict[str, Project] = {}
        all_names: Set[str] = set()
        for key, val in tree.items():
            all_names.add(key)
            if isinstance(val, Table):
                tables[key] = val
                val._project_url = path
            else:
                folders[key] = _recursive_create_project(f"{base_path}/{key}", val)
        # Any folder that hasn't been added as something else means there was
        # another object there that we could materialize.
        virtual_folders = (
            [key for key in additional_folders if key not in all_names]
            if additional_folders
            else []
        )
        ret = Project(
            root_base_path=path,
            base_path=base_path,
            tables=tables,
            folders=folders,
            virtual_folders=virtual_folders,
            fetch_minimal_table_metadata=_fetch_minimal_table_metadata,
            fetch_table_metadata=fetch_table_metadata,
            fetch_samples=fetch_samples,
        )
        if _eager:
            _recursive_eager_materialize(ret)
        return ret

    return _recursive_create_project(path, project_tree, folder.folders)



"""Vendored from xarray
https://github.com/pydata/xarray/blob/f4e0523be0ce8babaa8eff38365e5308b1fdb76b/xarray/core/formatting_html.py
"""

from __future__ import annotations

import uuid
from functools import partial
from html import escape
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional

from xarray.core.options import _get_boolean_with_default

from fused_batch._formatter.common import format_number, icon, load_static_files
from fused_batch._formatter.formatter_dataset import tables_section
from fused_batch._options import options as OPTIONS

if TYPE_CHECKING:
    from fused_batch._project import Project


def collapsible_section(
    name,
    inline_details: str = "",
    details: str = "",
    n_items: Optional[int] = None,
    enabled: bool = True,
    collapsed: bool = False,
):
    # "unique" id to expand/collapse the section
    data_id = "section-" + str(uuid.uuid4())

    has_items = n_items is not None and n_items
    n_items_span = "" if n_items is None else f" <span>({n_items})</span>"
    enabled_str = "" if enabled and has_items else "disabled"
    collapsed_str = "" if collapsed or not has_items else "checked"
    tip = " title='Expand/collapse section'" if enabled else ""

    return (
        f"<input id='{data_id}' class='xr-section-summary-in' "
        f"type='checkbox' {enabled_str} {collapsed_str}>"
        f"<label for='{data_id}' class='xr-section-summary' {tip}>"
        f"{name}:{n_items_span}</label>"
        f"<div class='xr-section-inline-details'>{inline_details}</div>"
        f"<div class='xr-section-details'>{details}</div>"
    )


def _mapping_section(
    mapping,
    name: str,
    details_func: Callable[[Any], str],
    max_items_collapse,
    expand_option_name,
    enabled: bool = True,
):
    n_items = len(mapping)
    expanded = _get_boolean_with_default(
        expand_option_name, n_items < max_items_collapse
    )
    collapsed = not expanded

    return collapsible_section(
        name,
        details=details_func(mapping),
        n_items=n_items,
        enabled=enabled,
        collapsed=collapsed,
    )


def summarize_folder(
    folder_name: str,
    folder: Project,
    *,
    is_index: bool = False,
):
    cssclass_idx = " class='xr-has-index'" if is_index else ""

    tables_folders_str = f"tables:{format_number(len(folder.tables))} folders:{format_number(len(folder.folders))}"

    # "unique" ids required to expand/collapse subsections
    attrs_id = "attrs-" + str(uuid.uuid4())
    data_id = "data-" + str(uuid.uuid4())
    chunk_repr_disabled = "disabled" if True else ""

    attrs_ul = f"""
    <ul style="margin-top: 0; margin-bottom: 0px;">
    {fused_project_repr(folder)}
    </ul>
    """
    data_repr = ""

    attrs_icon = icon("icon-file-text2")
    code_icon = icon("icon-code")

    return (
        f"<div class='xr-var-name'><span{cssclass_idx}>{folder_name}</span></div>"
        f"<div class='xr-var-dims'>{tables_folders_str}</div>"
        f"<div class='xr-var-dtype'></div>"
        f"<div class='xr-var-preview xr-preview'></div>"
        f"<input id='{attrs_id}' class='xr-var-attrs-in' "
        f"type='checkbox'>"
        f"<label for='{attrs_id}' title='Show/Hide attributes'>"
        f"{attrs_icon}</label>"
        f"<input id='{data_id}' class='xr-var-data-in' type='checkbox' {chunk_repr_disabled}>"
        f"<label for='{data_id}' title='Show/Hide data repr'>"
        f"{code_icon}</label>"
        f"<div class='xr-var-attrs'>{attrs_ul}</div>"
        f"<div class='xr-var-data'>{data_repr}</div>"
    )


def summarize_folders(folders: Dict[str, Project]):
    li_items = []
    for folder_name, folder in folders.items():
        li_content = summarize_folder(
            folder_name=folder_name,
            folder=folder,
            is_index=False,
        )
        li_items.append(f"<li class='xr-var-item'>{li_content}</li>")

    vars_li = "".join(li_items)

    return f"<ul class='xr-var-list'>{vars_li}</ul>"


folder_section = partial(
    _mapping_section,
    name="Folders",
    details_func=summarize_folders,
    max_items_collapse=25,
    expand_option_name="display_expand_coords",
)


def summarize_virtual_folders(folders: List[str]):
    li_items = []

    attrs_icon = icon("icon-file-text2")
    code_icon = icon("icon-code")

    for folder_name in folders:
        li_content = (
            f"<div class='xr-var-name'><span>{folder_name}</span></div>"
            f"<div class='xr-var-dims'></div>"
            f"<div class='xr-var-dtype'></div>"
            f"<div class='xr-var-preview xr-preview'></div>"
            f"<input class='xr-var-attrs-in' type='checkbox' disabled>"
            f"<label title='Show/Hide attributes'>"
            f"{attrs_icon}</label>"
            f"<input class='xr-var-data-in' type='checkbox' disabled>"
            f"<label title='Show/Hide data repr'>"
            f"{code_icon}</label>"
            f"<div class='xr-var-attrs'></div>"
            f"<div class='xr-var-data'></div>"
        )
        li_items.append(f"<li class='xr-var-item'>{li_content}</li>")

    vars_li = "".join(li_items)

    return f"<ul class='xr-var-list'>{vars_li}</ul>"


virtual_folder_section = partial(
    _mapping_section,
    name="Virtual Folders",
    details_func=summarize_virtual_folders,
    max_items_collapse=25,
    expand_option_name="display_expand_coords",
)


def _obj_repr(obj, header_components, sections):
    """Return HTML repr of an xarray object.

    If CSS is not injected (untrusted notebook), fallback to the plain text repr.

    """
    header = f"<div class='xr-header'>{''.join(h for h in header_components)}</div>"
    sections = "".join(f"<li class='xr-section-item'>{s}</li>" for s in sections)

    icons_svg, css_style = load_static_files()
    return (
        "<div>"
        f"{icons_svg}<style>{css_style}</style>"
        f"<pre class='xr-text-repr-fallback'>{escape(repr(obj))}</pre>"
        "<div class='xr-wrap' style='display:none'>"
        f"{header}"
        f"<ul class='xr-sections'>{sections}</ul>"
        "</div>"
        "</div>"
    )


def fused_project_repr(
    project: Project,
) -> str:
    obj_type = f"fused_batch.Project: {project.base_path}"
    header_components = [f"<div class='xr-obj-type'>{escape(obj_type)}</div>"]

    if OPTIONS.show.materialize_virtual_folders:
        project._materialize_all_virtual_folders()

    sections: List[str] = []

    if project.tables:
        sections.append(tables_section(project.tables))
    if project.folders:
        sections.append(folder_section(project.folders))
    if project.virtual_folders:
        sections.append(virtual_folder_section(project.virtual_folders))

    if not sections:
        # There is nothing whatsoever in the project (totally empty).
        # Display something so it's clear there's 0 items.
        # This can occur if the user opens a project in lazy mode or
        # using a path that doesn't exist.
        sections.append(tables_section({}))

    return _obj_repr(project, header_components, sections)



import json
import os
import shutil
import tempfile
import zipfile
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

from fused._options import options as OPTIONS
from fused.models.base import FusedBaseModel

from .udf import structure_params


class CustomJobConfig(FusedBaseModel):
    name: Optional[str] = None
    steps: List[Dict[str, Any]] = []  # Support steps of flexible structure
    metadata: Optional[Dict[str, Any]] = None
    _validate_version: bool = True


class MetaJson(FusedBaseModel):
    job_config: CustomJobConfig


def create_directory_and_zip(
    path: str, how: Literal["zip", "local"], files: Dict, overwrite: bool = False
):
    def _create_files_in_directory(path_obj, files, tmpdir=False):
        # Create dir
        if not path_obj.exists():
            path_obj.mkdir()

        # Write files
        for filename, file_contents in files.items():
            file_path = path_obj / filename
            with open(file_path, "w") as file:
                if isinstance(file_contents, Dict):
                    # Prettify JSON files
                    file_contents = json.dumps(
                        json.loads(json.dumps(file_contents)), indent=4, sort_keys=True
                    )
                file.write(file_contents)
        return path_obj

    def _write_files_to_zip(path_obj, path_obj_with_files):
        with zipfile.ZipFile(path_obj, "w", zipfile.ZIP_DEFLATED) as zipf:
            for root, _, files in os.walk(path_obj_with_files):
                for file in files:
                    source_filepath = Path(root) / file
                    relative_path = os.path.relpath(
                        source_filepath, path_obj_with_files
                    )

                    zipf.write(source_filepath, relative_path)

    if isinstance(path, (str, Path)):
        path_obj = Path(path)
        base_name = path_obj.stem

        if path_obj.exists():
            if not overwrite:
                raise FileExistsError(f"Object {path} already exists.")
            else:
                if path_obj.is_dir():
                    shutil.rmtree(path_obj)
                elif path_obj.is_file():
                    path_obj.unlink()
    else:
        path_obj = path
        base_name = "udf"

    if how == "zip":
        with tempfile.TemporaryDirectory(dir=str(OPTIONS.temp_directory)) as tmpdir:
            tmp_path_obj = Path(tmpdir) / base_name
            path_obj_with_files = _create_files_in_directory(
                tmp_path_obj, files, tmpdir=True
            )
            _write_files_to_zip(path_obj, path_obj_with_files)
    elif how == "local":
        if isinstance(path_obj, Path) and path_obj.suffix:
            # TODO: Warning, not error
            raise ValueError(
                "To export with `how='local'`, path must be a directory, not a file."
            )

        _create_files_in_directory(path_obj, files)
    else:
        raise NotImplementedError(f"{how=} is not implemented")


def generate_readme(job):
    """Structure text for README.md."""
    job_names = [f"job_{step.udf.name}" for step in job.steps]

    jobs = []
    for step in job.steps:
        jobs.append(
            f"job_{step.udf.name} = {step.udf.name}({structure_params(step._generate_job_params())})"
        )

    str_udf_imports = "\n".join(
        [f"from udf_{step.udf.name} import {step.udf.name}" for step in job.steps]
    )
    str_run_local = "job.run_local(file_id=0, chunk_id=0)"
    str_run_remote = "job.run_remote(output_table='output_table_name')"
    str_job = "\n".join(jobs)
    str_multijob = f"job = fused.experimental.job([{', '.join(job_names)}])"
    src = f"""
# Fused Multi-Step Job

## Get started
```python
# Import UDFs
{str_udf_imports}

# Instantiate individual jobs
{str_job}

# Instantiate multi-step job
{str_multijob}

# Run locally
{str_run_local}

# Run remotely
{str_run_remote}
```
"""
    return src


def generate_meta_json(job):
    """Structure json for `meta.json`."""
    job_config = job.model_dump()
    steps = []
    for step in job_config["steps"]:
        step["udf"]["source"] = f"{step['udf']['name']}.py"
        step["udf"]["metadata"] = {
            **(step["udf"]["metadata"] or {}),
            "fused:slug": step["udf"]["name"],  # METADATA_FUSED_SLUG
            "fused:name": step["udf"]["name"],  # METADATA_FIELD_NAME
        }
        del step["udf"]["code"]
        for header in step["udf"]["headers"]:
            # If there's no source file, introduce source file name
            if not header["source_file"]:
                header["source_file"] = header["module_name"] + ".py"  # virtual

            # Set relative path to sibling file.
            if not header["source_file"].startswith(("http", "https", "www")):
                header["source_file"] = header["module_name"] + ".py"
            # Remove `source_code` field.
            del header["source_code"]
        steps.append(step)

    job_config["steps"] = steps

    # Structure JSON
    return MetaJson(job_config=job_config).model_dump_json()



import sys
from functools import lru_cache
from typing import Optional, Union
from uuid import uuid4

from fused._options import options as OPTIONS

if sys.version_info >= (3, 9):
    from importlib.resources import files
else:
    from importlib_resources import files

STATIC_FILES = (
    ("fused._static.html", "icons-svg-inline.html"),
    ("fused._static.css", "style.css"),
)


def format_number(num: Union[float, int, None]) -> str:
    if OPTIONS.show.format_numbers is not False:
        # Note: We do not localize here (using {num:n}) because by default
        # there is no locale set, so this would do nothing.
        # If the locale thing is fixed, then this also needs to replace the
        # commas in a smarter way.
        text = f"{num:,}"
        text = text.replace(",", '<span style="user-select: none">,</span>')
        return text
    else:
        return f"{num}"


# Note: remove this lru_cache if changing the CSS
@lru_cache(None)
def load_static_files():
    """Lazily load the resource files into memory the first time they are needed"""
    return [
        files(package).joinpath(resource).read_text(encoding="utf-8")
        for package, resource in STATIC_FILES
    ]


def icon(icon_name):
    # icon_name should be defined in xarray/static/html/icon-svg-inline.html
    return (
        "<svg class='icon xr-{0}'>"
        "<use xlink:href='#{0}'>"
        "</use>"
        "</svg>".format(icon_name)
    )


def copyable_text(text: Optional[str], *, show_text: bool = True) -> str:
    """Returns an HTML fragment for a copyable text block"""
    if text is None:
        return ""
    id = f"copyable-{uuid4()}"
    js_fragment = f"var el = document.getElementById('{id}'); navigator.clipboard.writeText(el.innerText); var el2 = document.getElementById('{id}-copied'); el2.classList.add('fused-copy-fade-inout'); setTimeout(function(){{el2.classList.remove('fused-copy-fade-inout');}}, 1600);"
    copy_icon = icon("icon-copy")
    copy = f'<a onclick="javascript:{js_fragment}" class="fused-copy-button" title="Copy to clipboard">{copy_icon}</a>'
    code_style = 'style="display: none;"' if not show_text else ""
    copy_indicator = f'<span id="{id}-copied" class="fused-copy-hidden">Copied</span>'
    code_fragment = f'<code id="{id}" {code_style}>{text}</code>'
    return f"{code_fragment} {copy} {copy_indicator}"



import warnings
from operator import attrgetter
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    StrictBool,
    StrictStr,
    field_validator,
)

from ._global_api import reset_api
from .warnings import FusedIgnoredWarning, FusedImportWarning

DEV_DEFAULT_BASE_URL = "http://localhost:8783/v1"
PROD_DEFAULT_BASE_URL = "https://www.fused.io/server/v1"
PREVIEW_DEFAULT_BASE_URL = "https://www-preview.fused.io/server/v1"
STAGING_DEFAULT_BASE_URL = "https://www-staging.fused.io/server/v1"


OPTIONS_PATH = Path("~/.fused/settings.toml").expanduser()


class OptionsBaseModel(BaseModel):
    def __dir__(self) -> List[str]:
        # Provide method name lookup and completion. Only provide 'public'
        # methods.
        # This enables autocompletion
        # Pydantic methods to remove in __dir__
        PYDANTIC_METHODS = {
            "Config",
            "construct",
            "copy",
            "from_orm",
            "json",
            "parse_file",
            "parse_obj",
            "schema",
            "schema_json",
            "update_forward_refs",
            "validate",
            "model_validate",
            "model_dump_json",
        }

        normal_dir = {
            name
            for name in dir(type(self))
            if (not name.startswith("_") and name not in PYDANTIC_METHODS)
        }
        pydantic_fields = set(self.model_fields.keys())
        return sorted(normal_dir | pydantic_fields)

    def _repr_html_(self) -> str:
        # Circular import because the repr needs the options
        from fused_batch._formatter.formatter_options import fused_options_repr

        return fused_options_repr(self)


class OpenOptions(OptionsBaseModel):
    """Options for opening tables and projects."""

    fetch_samples: Optional[StrictBool] = None
    """Whether to automatically fetch samples when opening tables"""

    fetch_table_metadata: Optional[StrictBool] = True
    """Whether to automatically fetch table metadata when opening projects"""

    fetch_minimal_table_metadata: Optional[StrictBool] = True
    """Whether to fetch only a minimal set of table metadata when opening projects"""

    auto_refresh_project: StrictBool = True
    """Automatically refresh project objects when accessing a key that is not present locally"""


class ShowOptions(OptionsBaseModel):
    """Options for showing debug information"""

    open_browser: Optional[StrictBool] = None
    """Whether to open a local browser window for debug information"""
    show_widget: Optional[StrictBool] = None
    """Whether to show debug information in an IPython widget"""

    format_numbers: Optional[StrictBool] = None
    """Whether to format numbers in object reprs"""

    materialize_virtual_folders: StrictBool = True
    """Whether to automatically materialize virtual project folders in reprs"""


class CacheOptions(OptionsBaseModel):
    """Options for caching samples"""

    enable: StrictBool = True
    """Whether to enable caching"""


class Options(OptionsBaseModel):
    base_url: str = PROD_DEFAULT_BASE_URL
    """Fused API endpoint"""

    open: OpenOptions = Field(default_factory=OpenOptions)
    """Options for `fused.open_table` and `fused.open_project`."""
    show: ShowOptions = Field(default_factory=ShowOptions)
    """Options for object reprs and how data are shown for debugging."""
    cache: CacheOptions = Field(default_factory=CacheOptions)
    """Options for caching data fused-py can retrieve, such as
    the sample for `run_local`."""

    max_workers: int = 16
    """Maximum number of threads, when multithreading requests"""

    request_timeout: Union[Tuple[float, float], float, None] = 120
    """Request timeout for the Fused service

    May be set to a tuple of connection timeout and read timeout"""

    realtime_client_id: Optional[StrictStr] = None
    """Client ID for realtime service."""

    save_user_settings: StrictBool = True
    """Save per-user settings such as credentials and environment IDs."""

    default_udf_run_engine: Optional[StrictStr] = None
    """Default engine to run UDFs, one of: `local`, `realtime`, `batch`."""

    default_validate_imports: StrictBool = False
    """Default for whether to validate imports in UDFs before `run_local`,
    `run_batch`."""

    prompt_to_login: StrictBool = False
    """Automatically prompt the user to login when importing Fused."""

    @field_validator("base_url")
    @classmethod
    def _validate_base_url(cls, v):
        reset_api()
        return v

    def save(self):
        """Save Fused options to `~/.fused/settings.toml`. They will be automatically
        reloaded the next time fused-py is imported.
        """
        try:
            import rtoml

            OPTIONS_PATH.parent.mkdir(parents=True, exist_ok=True)
            # None (null) will not be serialized correctly in toml, so exclude it.
            # Any option which can be None should be None by default. Some open options
            # don't do this; should be updated to not be Optional.
            rtoml.dump(self.dict(exclude_none=True), OPTIONS_PATH)
        except ImportError:
            warnings.warn(
                "rtoml is not installed so options are not saved", FusedImportWarning
            )

    def _to_toml(self) -> str:
        try:
            import rtoml

            return rtoml.dumps(self.dict(exclude_none=True))
        except ImportError:
            warnings.warn(
                "rtoml is not installed so options are not saved", FusedImportWarning
            )

    model_config = ConfigDict(validate_assignment=True)


def _load_options():
    if OPTIONS_PATH.exists():
        try:
            import rtoml

            return Options.model_validate(rtoml.load(OPTIONS_PATH))
        except:  # noqa E722
            warnings.warn(
                f"Settings file {OPTIONS_PATH} exists but could not be loaded.",
                FusedIgnoredWarning,
            )

    return Options()


options = _load_options()
"""List global configuration options.

This object contains a set of configuration options that control global behavior of the library. This object can be used to modify the options.

Examples:
    Change the `request_timeout` option from its default value to 120 seconds:
    ```py
    fused.options.request_timeout = 120
    ```
"""


def set_option(option_name: str, option_value: Any):
    """
    Sets the value of a configuration option.

    This function updates the global `options` object with a new value for a specified option.
    It supports setting values for nested options using dot notation. For example, if the
    `options` object has a nested structure, you can set a value for a nested attribute
    by specifying the option name in the form "parent.child".

    Args:
        option_name: A string specifying the name of the option to set. This can be a simple
                     attribute name or a dot-separated path for nested attributes.
        option_value: The new value to set for the specified option. This can be of any type
                      that is compatible with the attribute being set.

    Raises:
        AttributeError: If the specified attribute path is not valid, either because a part
                        of the path does not exist or the final attribute cannot be set with
                        the provided value.

    Examples:
        Set the `request_timeout` top-level option to 120 seconds:
        ```python
        set_option('request_timeout', 120)
        ```


    """
    # Set a recursive option name
    # From https://stackoverflow.com/a/65355793 and comments
    if "." in option_name:
        split = option_name.split(".")
        parent = ".".join(split[:-1])
        base_name = split[-1]
        setattr(attrgetter(parent)(options), base_name, option_value)
        return

    setattr(options, option_name, option_value)


def _env(environment_name: str):
    """Set the environment."""
    if environment_name == "dev":
        _env = DEV_DEFAULT_BASE_URL
    elif environment_name == "stg":
        _env = STAGING_DEFAULT_BASE_URL
    elif environment_name == "prod":
        _env = PROD_DEFAULT_BASE_URL
    elif environment_name == "pre":
        _env = PREVIEW_DEFAULT_BASE_URL
    else:
        raise ValueError("Available options are `dev`, `stg`, `prod`, and `pre`.")

    setattr(options, "base_url", _env)



from typing import List

from fused_batch._formatter.udf import fused_registry_repr


class UdfRegistry(dict):
    def _repr_html_(self):
        return fused_registry_repr(self)

    def __getattribute__(self, key):
        try:
            return super().__getattribute__(key)
        except AttributeError:
            try:
                return self[key]
            # IPython's _repr_html_ requires AttributeError instead of KeyError.
            except KeyError:
                raise AttributeError(
                    f"object of type {type(self).__name__} has no attribute {key}"
                ) from None

    def __dir__(self) -> List[str]:
        """Provide method name lookup and completion."""
        extra_attrs = set(self.keys())
        return sorted(extra_attrs)
